import pybnesian as pbn
import pyarrow as pa


def test_type_equality():
    # 
    # Test single vector types
    # 

    het_single = pbn.HeterogeneousBN([pbn.CKDEType(), pbn.LinearGaussianCPDType()], ["a", "b", "c", "d"])
    het2_single = pbn.HeterogeneousBN([pbn.CKDEType(), pbn.LinearGaussianCPDType()], ["a", "b", "c", "d"])

    assert het_single.type() == het2_single.type()

    het3_single = pbn.HeterogeneousBN([pbn.LinearGaussianCPDType(), pbn.CKDEType()], ["a", "b", "c", "d"])
    
    assert het_single.type() != het3_single.type()

    # 
    # Test a single vector type for each data type
    # 

    het_dt = pbn.HeterogeneousBN({
        pa.float64(): [pbn.CKDEType(), pbn.LinearGaussianCPDType()],
        pa.float32(): [pbn.CKDEType(), pbn.LinearGaussianCPDType()],
        pa.dictionary(pa.int8(), pa.string()): [pbn.DiscreteFactorType()]
    }, ["a", "b", "c", "d"])

    het2_dt = pbn.HeterogeneousBN({
        pa.dictionary(pa.int8(), pa.string()): [pbn.DiscreteFactorType()],
        pa.float32(): [pbn.CKDEType(), pbn.LinearGaussianCPDType()],
        pa.float64(): [pbn.CKDEType(), pbn.LinearGaussianCPDType()]
    }, ["a", "b", "c", "d"])
    
    # The order of the set is not relevant
    assert het_dt.type() == het2_dt.type()

    het3_dt = pbn.HeterogeneousBN({
        pa.dictionary(pa.int8(), pa.string()): [pbn.DiscreteFactorType()],
        pa.float32(): [pbn.LinearGaussianCPDType(), pbn.CKDEType()],
        pa.float64(): [pbn.CKDEType(), pbn.LinearGaussianCPDType()]
    }, ["a", "b", "c", "d"])

    # The order of the default FactorTypes is relevant
    assert het_dt.type() != het3_dt.type()
    
    # 
    # Compare single vector and multi vector FactorTypes

    het_single = pbn.HeterogeneousBN([pbn.CKDEType(), pbn.LinearGaussianCPDType()], ["a", "b", "c", "d"])
    het_dt = pbn.HeterogeneousBN({
        pa.float64(): [pbn.CKDEType(), pbn.LinearGaussianCPDType()]
    }, ["a", "b", "c", "d"])

    assert het_single.type() != het_dt.type()



import pytest
import pybnesian as pbn
import util_test

SIZE = 10000
df = util_test.generate_normal_data(SIZE)

def test_create():
    arcs = pbn.ArcOperatorSet()
    node_type = pbn.ChangeNodeTypeSet()
    pool = pbn.OperatorPool([arcs, node_type])

    with pytest.raises(ValueError) as ex:
        pool = pbn.OperatorPool([])
    assert "cannot be empty" in str(ex.value)

def test_find_max():
    spbn = pbn.SemiparametricBN(['a', 'b', 'c', 'd'])
    cv = pbn.CVLikelihood(df)
    arcs = pbn.ArcOperatorSet()
    node_type = pbn.ChangeNodeTypeSet()
    
    arcs.cache_scores(spbn, cv)
    spbn.set_unknown_node_types(df)
    node_type.cache_scores(spbn, cv)
    
    arcs_max = arcs.find_max(spbn)
    node_max = node_type.find_max(spbn)

    pool = pbn.OperatorPool([arcs, node_type])
    pool.cache_scores(spbn, cv)

    op_combined = pool.find_max(spbn)

    if arcs_max.delta() >= node_max.delta():
        assert op_combined == arcs_max
    else:
        assert op_combined == node_max

    


import pytest
import numpy as np
import pyarrow as pa
import pandas as pd
import pybnesian as pbn
from scipy.stats import gaussian_kde
from scipy.stats import norm
from scipy.stats import multivariate_normal as mvn
from scipy.special import logsumexp

import util_test

SIZE = 10000
SMALL_SIZE = 10
TEST_SIZE = 50
df = util_test.generate_normal_data(SIZE, seed=0)
df_small = util_test.generate_normal_data(SMALL_SIZE, seed=0)
df_float = df.astype('float32')
df_small_float = df_small.astype('float32')

def test_variable():
    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.CKDE(variable, evidence)
        assert cpd.variable() == variable

def test_evidence():
    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.CKDE(variable, evidence)
        assert cpd.evidence() == evidence

def test_kde_data_type():
    k = pbn.CKDE("a", [])

    with pytest.raises(ValueError) as ex:
        k.data_type()
    "CKDE factor not fitted" in str(ex.value)

    k.fit(df)
    assert k.data_type() == pa.float64()
    k.fit(df_float)
    assert k.data_type() == pa.float32()

def test_ckde_kde_joint():
    def _test_ckde_kde_joint_iter(variable, evidence, _df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)
        kde_joint = cpd.kde_joint
        kde_joint().bandwidth = np.eye(len(evidence) + 1)
        assert np.all(cpd.kde_joint().bandwidth == np.eye(len(evidence) + 1)), "kde_joint do not return a reference to the KDE joint, but a copy."

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_kde_joint_iter(variable, evidence, df)
        _test_ckde_kde_joint_iter(variable, evidence, df_float)

def test_ckde_kde_marg():
    def _test_ckde_kde_marg_iter(variable, evidence, _df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)
        kde_marg = cpd.kde_marg

        if evidence:
            assert kde_marg().fitted()
            kde_marg().bandwidth = np.eye(len(evidence))
            assert np.all(cpd.kde_marg().bandwidth == np.eye(len(evidence))), "kde_marg do not return a reference to the KDE joint, but a copy."
        else:
            # kde_marg contains garbage if there is no evidence
            pass

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_kde_marg_iter(variable, evidence, df)
        _test_ckde_kde_marg_iter(variable, evidence, df_float)

def test_ckde_fit():
    def _test_ckde_fit(variables, _df, instances):
        npdata = _df.loc[:, variables].to_numpy()
        scipy_kde = gaussian_kde(npdata[:instances, :].T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        cpd = pbn.CKDE(variable, evidence)
        assert not cpd.fitted()
        cpd.fit(_df.iloc[:instances])
        assert cpd.fitted()

        kde_joint = cpd.kde_joint
        assert np.all(np.isclose(kde_joint().bandwidth, scipy_kde.covariance))
        
        if evidence:
            kde_marg = cpd.kde_marg
            assert np.all(np.isclose(kde_marg().bandwidth, scipy_kde.covariance[1:,1:]))
        
        assert cpd.num_instances() == instances

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        variables = [variable] + evidence
        for instances in [50, 1000, 10000]:            
            _test_ckde_fit(variables, df, instances)
            _test_ckde_fit(variables, df_float, instances)

def test_ckde_fit_null():
    def _test_ckde_fit_null(variable, evidence, variables, _df, instances):
        cpd = pbn.CKDE(variable, evidence)
        assert not cpd.fitted()
        cpd.fit(_df.iloc[:instances])
        assert cpd.fitted()

        npdata = _df.loc[:, variables].to_numpy()
        npdata_instances = npdata[:instances,:]
        nan_rows = np.any(np.isnan(npdata_instances), axis=1)
        npdata_no_null = npdata_instances[~nan_rows,:]
        scipy_kde = gaussian_kde(npdata_no_null.T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        kde_joint = cpd.kde_joint
        assert np.all(np.isclose(kde_joint().bandwidth, scipy_kde.covariance))
        
        if evidence:
            kde_marg = cpd.kde_marg
            assert np.all(np.isclose(kde_marg().bandwidth, scipy_kde.covariance[1:,1:]))

        assert cpd.num_instances() == scipy_kde.n

    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        variables = [variable] + evidence
        for instances in [50, 1000, 10000]:            
            _test_ckde_fit_null(variable, evidence, variables, df, instances)
            _test_ckde_fit_null(variable, evidence, variables, df_float, instances)

def train_scipy_ckde(data, variable, evidence):
    variables = [variable] + evidence
    npdata_joint = data.loc[:, variables].to_numpy()
    npdata_marg = data.loc[:, evidence].to_numpy()

    nan_rows = np.any(np.isnan(npdata_joint), axis=1)

    scipy_kde_joint = gaussian_kde(npdata_joint[~nan_rows,:].T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())
    if evidence:
        scipy_kde_marg = gaussian_kde(npdata_marg[~nan_rows,:].T, bw_method=scipy_kde_joint.covariance_factor())
    else:
        scipy_kde_marg = None

    return scipy_kde_joint, scipy_kde_marg

def scipy_ckde_logpdf(test_data, joint_kde, marg_kde, variable, evidence):
    variables = [variable] + evidence
    test_data_joint = test_data.loc[:, variables].to_numpy()
    test_data_marg = test_data.loc[:, evidence].to_numpy()

    nan_rows = np.any(np.isnan(test_data_joint), axis=1)

    if np.all(test_data.dtypes == "float32"):
        result = np.full(test_data.shape[0], np.nan, dtype=np.float32)
    else:
        result = np.full(test_data.shape[0], np.nan, dtype=np.float64)

    if evidence:
        result[~nan_rows] = joint_kde.logpdf(test_data_joint[~nan_rows,:].T) - marg_kde.logpdf(test_data_marg[~nan_rows,:].T)
    else:
        result[~nan_rows] = joint_kde.logpdf(test_data_joint[~nan_rows,:].T)

    return result

def scipy_ckde_cdf(test_data, joint_kde, marg_kde, variable, evidence):
    variables = [variable] + evidence
    test_data_joint = test_data.loc[:, variables].to_numpy()
    test_data_marg = test_data.loc[:, evidence].to_numpy()

    nan_rows = np.any(np.isnan(test_data_joint), axis=1)

    if np.all(test_data.dtypes == "float32"):
        result = np.full(test_data.shape[0], np.nan, dtype=np.float32)
    else:
        result = np.full(test_data.shape[0], np.nan, dtype=np.float64)

    total_w = np.empty((joint_kde.n, test_data_joint.shape[0]))
    conditional_mean = np.empty((joint_kde.n, test_data_joint.shape[0]))
    total_cdf = np.empty((joint_kde.n, test_data_joint.shape[0]))

    if evidence:
        bandwidth = joint_kde.covariance
        cond_var = bandwidth[0,0] - bandwidth[0, 1:].dot(np.linalg.inv(bandwidth[1:, 1:])).dot(bandwidth[1:, 0])
        for test_index in np.where(~np.any(np.isnan(test_data_joint), axis=1))[0]:
            w = mvn.logpdf(marg_kde.dataset.T, mean=test_data_marg[test_index,:], cov=marg_kde.covariance)
            w = np.exp(w)
            total_w[:, test_index] = w

            evidence_diff = test_data_marg[test_index,:] - joint_kde.dataset[1:,:].T
            cond_mean = joint_kde.dataset[0,:] + bandwidth[0,1:].dot(np.linalg.inv(bandwidth[1:,1:])).dot(evidence_diff.T)

            conditional_mean[:, test_index] = cond_mean
            total_cdf[:, test_index] = norm.cdf(test_data_joint[test_index,0], cond_mean, np.sqrt(cond_var))

            result[test_index] = np.dot(w, norm.cdf(test_data_joint[test_index,0], cond_mean, np.sqrt(cond_var)))

        result /= np.sum(total_w, axis=0)

    else:
        cdf = norm.cdf(test_data_joint[~nan_rows], joint_kde.dataset, np.sqrt(joint_kde.covariance[0,0]))
        result[~nan_rows] = np.sum((1 / joint_kde.n) * cdf, axis=1)

    return result

def test_ckde_logl():
    def _test_ckde_logl(variable, evidence, _df, _test_df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)
        scipy_kde_joint, scipy_kde_marg = train_scipy_ckde(_df, variable, evidence)

        logl = cpd.logl(_test_df)
        scipy = scipy_ckde_logpdf(_test_df, scipy_kde_joint, scipy_kde_marg, variable, evidence)

        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(logl, scipy, atol=0.0005))
        else:
            assert np.all(np.isclose(logl, scipy))
    
    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_logl(variable, evidence, df, test_df)
        _test_ckde_logl(variable, evidence, df_small, test_df)
        _test_ckde_logl(variable, evidence, df_float, test_df_float)
        _test_ckde_logl(variable, evidence, df_small_float, test_df_float)

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.logl(test_df), cpd2.logl(test_df))), "Order of evidence changes logl() result."

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.logl(test_df_float), cpd2.logl(test_df_float), atol=0.0005)), "Order of evidence changes logl() result."

def test_ckde_logl_null():
    def _test_ckde_logl_null(variable, evidence, _df, _test_df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)

        scipy_kde_joint, scipy_kde_marg = train_scipy_ckde(_df, variable, evidence)

        logl = cpd.logl(_test_df)
        scipy = scipy_ckde_logpdf(_test_df, scipy_kde_joint, scipy_kde_marg, variable, evidence)

        if np.all(_test_df.dtypes == "float32"):
            assert np.all(np.isclose(logl, scipy, atol=0.0005, equal_nan=True))
        else:
            assert np.all(np.isclose(logl, scipy, equal_nan=True))

    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    np.random.seed(0)
    a_null = np.random.randint(0, TEST_SIZE, size=10)
    b_null = np.random.randint(0, TEST_SIZE, size=10)
    c_null = np.random.randint(0, TEST_SIZE, size=10)
    d_null = np.random.randint(0, TEST_SIZE, size=10)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = test_df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_logl_null(variable, evidence, df, df_null)
        _test_ckde_logl_null(variable, evidence, df_small, df_null)
        _test_ckde_logl_null(variable, evidence, df_float, df_null_float)
        _test_ckde_logl_null(variable, evidence, df_small_float, df_null_float)

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df)

    ll = cpd.logl(df_null)
    ll2 = cpd2.logl(df_null)
    assert np.all(np.isclose(ll, ll2, equal_nan=True)), "Order of evidence changes the position of nan values."

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df_float)

    ll = cpd.logl(df_null_float)
    ll2 = cpd2.logl(df_null_float)
    assert np.all(np.isclose(ll, ll2, equal_nan=True)), "Order of evidence changes the position of nan values."

def test_ckde_slogl():
    def _test_ckde_slogl(variable, evidence, _df, _test_df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)

        scipy_kde_joint, scipy_kde_marg = train_scipy_ckde(_df, variable, evidence)
        scipy_logl = scipy_ckde_logpdf(_test_df, scipy_kde_joint, scipy_kde_marg, variable, evidence)

        if np.all(_test_df.dtypes == "float32"):
            # Allow an error of 0.0005 for each training instance.
            assert np.isclose(cpd.slogl(_test_df), scipy_logl.sum(), atol=0.0005*_df.shape[0])
        else:
            assert np.isclose(cpd.slogl(_test_df), scipy_logl.sum())

    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_slogl(variable, evidence, df, test_df)
        _test_ckde_slogl(variable, evidence, df_small, test_df)
        _test_ckde_slogl(variable, evidence, df_float, test_df_float)
        _test_ckde_slogl(variable, evidence, df_small_float, test_df_float)

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.slogl(test_df), cpd2.slogl(test_df))), "Order of evidence changes slogl() result."

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.slogl(test_df_float), cpd2.slogl(test_df_float))), "Order of evidence changes slogl() result."

def test_ckde_slogl_null():
    def _test_ckde_slogl_null(variable, evidence, _df, _test_df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)

        scipy_kde_joint, scipy_kde_marg = train_scipy_ckde(_df, variable, evidence)
        scipy_logl = scipy_ckde_logpdf(_test_df, scipy_kde_joint, scipy_kde_marg, variable, evidence)

        if np.all(_test_df.dtypes == "float32"):
            # Allow an error of 0.0005 for each training instance.
            assert np.isclose(cpd.slogl(_test_df), np.nansum(scipy_logl), atol=0.0005*_df.shape[0])
        else:
            assert np.isclose(cpd.slogl(_test_df), np.nansum(scipy_logl))


    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    np.random.seed(0)
    a_null = np.random.randint(0, TEST_SIZE, size=10)
    b_null = np.random.randint(0, TEST_SIZE, size=10)
    c_null = np.random.randint(0, TEST_SIZE, size=10)
    d_null = np.random.randint(0, TEST_SIZE, size=10)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = test_df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_slogl_null(variable, evidence, df, df_null)
        _test_ckde_slogl_null(variable, evidence, df_small, df_null)
        _test_ckde_slogl_null(variable, evidence, df_float, df_null_float)
        _test_ckde_slogl_null(variable, evidence, df_small_float, df_null_float)


    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.slogl(df_null), cpd2.slogl(df_null))), "Order of evidence changes slogl() result."

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.slogl(df_null_float), cpd2.slogl(df_null_float))), "Order of evidence changes slogl() result."

def test_ckde_cdf():
    def _test_ckde_cdf(variable, evidence, _df, _test_df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)
        scipy_kde_joint, scipy_kde_marg = train_scipy_ckde(_df, variable, evidence)

        cdf = cpd.cdf(_test_df)
        scipy = scipy_ckde_cdf(_test_df, scipy_kde_joint, scipy_kde_marg, variable, evidence)

        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(cdf, scipy, atol=0.0005))
        else:
            assert np.all(np.isclose(cdf, scipy))
    
    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_cdf(variable, evidence, df, test_df)
        _test_ckde_cdf(variable, evidence, df_small, test_df)
        _test_ckde_cdf(variable, evidence, df_float, test_df_float)
        _test_ckde_cdf(variable, evidence, df_small_float, test_df_float)

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.cdf(test_df), cpd2.cdf(test_df))), "Order of evidence changes logl() result."

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.cdf(test_df_float), cpd2.cdf(test_df_float), atol=0.0005)), "Order of evidence changes logl() result."

def test_ckde_cdf_null():
    def _test_ckde_cdf_null(variable, evidence, _df, _test_df):
        cpd = pbn.CKDE(variable, evidence)
        cpd.fit(_df)

        scipy_kde_joint, scipy_kde_marg = train_scipy_ckde(_df, variable, evidence)

        cdf = cpd.cdf(_test_df)
        scipy = scipy_ckde_cdf(_test_df, scipy_kde_joint, scipy_kde_marg, variable, evidence)

        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(cdf, scipy, atol=0.0005, equal_nan=True))
        else:
            assert np.all(np.isclose(cdf, scipy, equal_nan=True))


    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    np.random.seed(0)
    a_null = np.random.randint(0, TEST_SIZE, size=10)
    b_null = np.random.randint(0, TEST_SIZE, size=10)
    c_null = np.random.randint(0, TEST_SIZE, size=10)
    d_null = np.random.randint(0, TEST_SIZE, size=10)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = test_df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        _test_ckde_cdf_null(variable, evidence, df, df_null)
        _test_ckde_cdf_null(variable, evidence, df_small, df_null)
        _test_ckde_cdf_null(variable, evidence, df_float, df_null_float)
        _test_ckde_cdf_null(variable, evidence, df_small_float, df_null_float)


    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.cdf(df_null), cpd2.cdf(df_null), equal_nan=True)), "Order of evidence changes cdf() result."

    cpd = pbn.CKDE('d', ['a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.CKDE('d', ['c', 'b', 'a'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.cdf(df_null_float), 
                             cpd2.cdf(df_null_float), 
                             atol=0.0005, equal_nan=True)), "Order of evidence changes cdf() result."

def test_ckde_sample():
    SAMPLE_SIZE = 1000

    cpd = pbn.CKDE('a', [])
    cpd.fit(df)
    
    sampled = cpd.sample(SAMPLE_SIZE, None, 0)

    assert sampled.type == pa.float64()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE
        
    cpd = pbn.CKDE('b', ['a'])
    cpd.fit(df)

    sampling_df = pd.DataFrame({'a': np.full((SAMPLE_SIZE,), 3.0)})
    sampled = cpd.sample(SAMPLE_SIZE, sampling_df, 0)

    assert sampled.type == pa.float64()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE
    
    cpd = pbn.CKDE('c', ['a', 'b'])
    cpd.fit(df)

    sampling_df = pd.DataFrame({'a': np.full((SAMPLE_SIZE,), 3.0),
                                'b': np.full((SAMPLE_SIZE,), 7.45)})
    sampled = cpd.sample(SAMPLE_SIZE, sampling_df, 0)

    assert sampled.type == pa.float64()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE

    cpd = pbn.CKDE('a', [])
    cpd.fit(df_float)
    
    sampled = cpd.sample(SAMPLE_SIZE, None, 0)

    assert sampled.type == pa.float32()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE
        
    cpd = pbn.CKDE('b', ['a'])
    cpd.fit(df_float)

    sampling_df = pd.DataFrame({'a': np.full((SAMPLE_SIZE,), 3.0, dtype=np.float32)})
    sampled = cpd.sample(SAMPLE_SIZE, sampling_df, 0)

    assert sampled.type == pa.float32()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE
    
    cpd = pbn.CKDE('c', ['a', 'b'])
    cpd.fit(df_float)

    sampling_df = pd.DataFrame({'a': np.full((SAMPLE_SIZE,), 3.0, dtype=np.float32),
                                'b': np.full((SAMPLE_SIZE,), 7.45, dtype=np.float32)})
    sampled = cpd.sample(SAMPLE_SIZE, sampling_df, 0)

    assert sampled.type == pa.float32()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE


import os
import conv_template


def expand_sources():

    sources = ['pybnesian/kde/opencl_kernels/KDE.cl.src']
    
    for source in sources:
        (base, _) = os.path.splitext(source)
        outstr = conv_template.process_file(source)
        with open(base, 'w') as fid:
            fid.write(outstr)


def copy_opencl_code():
    sources = ['pybnesian/kde/opencl_kernels/KDE.cl']

    # Split the CPP code because the MSVC only allow strings of a max size.
    # Error C2026: https://docs.microsoft.com/en-us/cpp/error-messages/compiler-errors-1/compiler-error-c2026?view=msvc-160
    MAX_LENGTH=16378
    code_str = ""
    for source in sources:
        code_str += '\n'
        with open(source) as f:
            source_code = f.read()
            code_str += source_code

    fragments = [code_str[i:(i + MAX_LENGTH)] for i in range(0, len(code_str), MAX_LENGTH)]

    cpp_code = \
    """#ifndef PYBNESIAN_OPENCL_OPENCL_CODE_HPP
#define PYBNESIAN_OPENCL_OPENCL_CODE_HPP

namespace opencl {
const std::string OPENCL_CODE = """

    for f in fragments:
        cpp_code += 'R"foo({})foo"'.format(f) + "\n"

    cpp_code += """;
}
#endif //PYBNESIAN_OPENCL_OPENCL_CODE_HPP"""

    with open('pybnesian/opencl/opencl_code.hpp', 'w') as f:
        f.write(cpp_code)


if __name__ == "__main__":
    expand_sources()
    copy_opencl_code()


import pytest
import pybnesian as pbn
from pybnesian import FactorType, Factor

def test_factor_type():
    lg1 = pbn.LinearGaussianCPD("a", [])
    lg2 = pbn.LinearGaussianCPD("b", ["a"])
    lg3 = pbn.LinearGaussianCPD("c", ["b", "a"])

    assert lg1.type() == pbn.LinearGaussianCPDType()
    assert lg1.type() == lg2.type()
    assert lg1.type() == lg3.type()
    assert lg2.type() == lg3.type()

    c1 = pbn.CKDE("a", [])
    c2 = pbn.CKDE("b", ["a"])
    c3 = pbn.CKDE("c", ["b", "a"])

    assert c1.type() == pbn.CKDEType()
    assert c1.type() == c2.type()
    assert c1.type() == c3.type()
    assert c2.type() == c3.type()

    d1 = pbn.DiscreteFactor("a", [])
    d2 = pbn.DiscreteFactor("b", ["a"])
    d3 = pbn.DiscreteFactor("c", ["b", "a"])

    assert d1.type() == pbn.DiscreteFactorType()
    assert d1.type() == d2.type()
    assert d1.type() == d3.type()
    assert d2.type() == d3.type()

    assert lg1.type() != c1.type()
    assert lg1.type() != d1.type()
    assert c1.type() != d1.type()

def test_new_factor_type():
    class A(FactorType):
        def __init__(self):
            FactorType.__init__(self)

    a1 = A()
    a2 = A()
    a3 = A()

    assert a1 == a2
    assert a1 == a3
    assert a2 == a3

    class B(FactorType):
        def __init__(self):
            FactorType.__init__(self)

    b1 = B()
    b2 = B()
    b3 = B()

    assert b1 == b2
    assert b1 == b3
    assert b2 == b3

    assert a1 != b1

def test_factor_defined_factor_type():
    class F_type(FactorType):
        def __init__(self):
            FactorType.__init__(self)

        def __str__(self):
            return "FType"

    class F(Factor):
        def __init__(self, variable, evidence):
            Factor.__init__(self, variable, evidence)

        def type(self):
            return F_type()

    f1 = F("a", [])
    f2 = F("b", ["a"])
    f3 = F("c", ["a", "b"])

    assert f1.type() == f2.type()
    assert f1.type() == f3.type()
    assert f2.type() == f3.type()

    assert str(f1.type()) == str(f2.type()) == str(f3.type()) == "FType"

    dummy_network = pbn.GaussianNetwork(["a", "b", "c", "d"])
    with pytest.raises(RuntimeError) as ex:
        f4 = f1.type().new_factor(dummy_network, "d", ["a", "b", "c"])
    assert 'Tried to call pure virtual function "FactorType::new_factor"' in str(ex.value)

    class G_type(FactorType):
        def __init__(self):
            FactorType.__init__(self)
            
        def new_factor(self, model, variable, evidence):
            return G(variable, evidence)

        def __str__(self):
            return "GType"

    class G(Factor):
        def __init__(self, variable, evidence):
            Factor.__init__(self, variable, evidence)

        def type(self):
            return G_type()

    g1 = G("a", [])
    g2 = G("b", ["a"])
    g3 = G("c", ["a", "b"])

    assert g1.type() == g2.type()
    assert g1.type() == g3.type()
    assert g2.type() == g3.type()

    assert f1.type() != g1.type()

    assert str(g1.type()) == str(g2.type()) == str(g3.type()) == "GType"

    g4 = g1.type().new_factor(dummy_network, "d", ["a", "b", "c"])

    assert g1.type() == g4.type()
    assert g4.variable() == "d"
    assert g4.evidence() == ["a", "b", "c"]


import pybnesian as pbn
from pybnesian import BayesianNetworkType, BayesianNetwork, ConditionalBayesianNetwork, GaussianNetwork,\
    SemiparametricBN, KDENetwork, DiscreteBN
import util_test

def test_bn_type():
    g1 = GaussianNetwork(["a", "b", "c", "d"])
    g2 = GaussianNetwork(["a", "b", "c", "d"])
    g3 = GaussianNetwork(["a", "b", "c", "d"])

    assert g1.type() == pbn.GaussianNetworkType()
    assert g1.type() == g2.type()
    assert g1.type() == g3.type()
    assert g2.type() == g3.type()

    s1 = SemiparametricBN(["a", "b", "c", "d"])
    s2 = SemiparametricBN(["a", "b", "c", "d"])
    s3 = SemiparametricBN(["a", "b", "c", "d"])

    assert s1.type() == pbn.SemiparametricBNType()
    assert s1.type() == s2.type()
    assert s1.type() == s3.type()
    assert s2.type() == s3.type()

    k1 = KDENetwork(["a", "b", "c", "d"])
    k2 = KDENetwork(["a", "b", "c", "d"])
    k3 = KDENetwork(["a", "b", "c", "d"])

    assert k1.type() == pbn.KDENetworkType()
    assert k1.type() == k2.type()
    assert k1.type() == k3.type()
    assert k2.type() == k3.type()

    d1 = DiscreteBN(["a", "b", "c", "d"])
    d2 = DiscreteBN(["a", "b", "c", "d"])
    d3 = DiscreteBN(["a", "b", "c", "d"])

    assert d1.type() == pbn.DiscreteBNType()
    assert d1.type() == d2.type()
    assert d1.type() == d3.type()
    assert d2.type() == d3.type()

    assert g1.type() != s1.type()
    assert g1.type() != k1.type()
    assert g1.type() != d1.type()
    assert s1.type() != k1.type()
    assert s1.type() != d1.type()
    assert k1.type() != d1.type()

def test_new_bn_type():
    class MyGaussianNetworkType(BayesianNetworkType):
        def __init__(self):
            BayesianNetworkType.__init__(self)

        def is_homogeneous(self):
            return True

        def can_have_arc(self, model, source, target):
            return source == "a"

    a1 = MyGaussianNetworkType()
    a2 = MyGaussianNetworkType()
    a3 = MyGaussianNetworkType()

    assert a1 == a2
    assert a1 == a3
    assert a2 == a3

    class MySemiparametricBNType(BayesianNetworkType):
        def __init__(self):
            BayesianNetworkType.__init__(self)
    
    b1 = MySemiparametricBNType()
    b2 = MySemiparametricBNType()
    b3 = MySemiparametricBNType()

    assert b1 == b2
    assert b1 == b3
    assert b2 == b3

    assert a1 != b1

    mybn = BayesianNetwork(a1, ["a", "b", "c", "d"])

    # This type omits the arcs that do not have "a" as source.
    assert mybn.can_add_arc("a", "b")
    assert not mybn.can_add_arc("b", "a")
    assert not mybn.can_add_arc("c", "d")


class MyRestrictedGaussianNetworkType(BayesianNetworkType):
    def __init__(self):
        BayesianNetworkType.__init__(self)

    def is_homogeneous(self):
        return True

    def default_node_type(self):
        return pbn.LinearGaussianCPDType()

    def can_have_arc(self, model, source, target):
        return source == "a"

    def __str__(self):
        return "MyRestrictedGaussianNetworkType"

class SpecificNetwork(BayesianNetwork):
    def __init__(self, variables, arcs=None):
        if arcs is None:
            BayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables)
        else:
            BayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, arcs)

class ConditionalSpecificNetwork(ConditionalBayesianNetwork):
    def __init__(self, variables, interface, arcs=None):
        if arcs is None:
            ConditionalBayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, interface)
        else:
            ConditionalBayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, interface, arcs)

def test_new_specific_bn_type():
    sp1 = SpecificNetwork(["a", "b", "c", "d"])
    sp2 = SpecificNetwork(["a", "b", "c", "d"], [("a", "b")])
    sp3 = SpecificNetwork(["a", "b", "c", "d"])

    assert sp1.type() == sp2.type()
    assert sp1.type() == sp3.type()
    assert sp2.type() == sp3.type()

    assert sp1.can_add_arc("a", "b")
    assert not sp1.can_add_arc("b", "a")
    assert not sp1.can_add_arc("c", "d")

    assert sp1.num_arcs() == sp3.num_arcs() == 0
    assert sp2.arcs() == [("a", "b")]

    df = util_test.generate_normal_data_indep(1000)
    bic = pbn.BIC(df)

    start = SpecificNetwork(["a", "b", "c", "d"])

    hc = pbn.GreedyHillClimbing()
    estimated = hc.estimate(pbn.ArcOperatorSet(), bic, start)
    assert estimated.type() == start.type()
    assert all([s == "a" for s, t in estimated.arcs()])

    # #######################
    # Conditional BN
    # #######################
 
    csp1 = ConditionalSpecificNetwork(["a", "b"], ["c", "d"])
    csp2 = ConditionalSpecificNetwork(["a", "b"], ["c", "d"], [("a", "b")])
    csp3 = ConditionalSpecificNetwork(["a", "b"], ["c", "d"])

    assert csp1.type() == csp2.type()
    assert csp1.type() == csp3.type()
    assert csp2.type() == csp3.type()

    assert csp1.can_add_arc("a", "b")
    assert not csp1.can_add_arc("b", "a")
    assert not csp1.can_add_arc("c", "d")

    assert csp1.num_arcs() == csp3.num_arcs() == 0
    assert csp2.arcs() == [("a", "b")]

    cstart = ConditionalSpecificNetwork(["a", "c"], ["b", "d"])

    hc = pbn.GreedyHillClimbing()
    cestimated = hc.estimate(pbn.ArcOperatorSet(), bic, cstart)
    assert cestimated.type() == cstart.type()
    assert all([s == "a" for s, t in cestimated.arcs()])



import pytest
import numpy as np
from scipy.stats import norm, gaussian_kde
import pybnesian as pbn
import util_test

SIZE = 1000
df = util_test.generate_normal_data(SIZE)

seed = 0

def numpy_local_score(node_type, data, variable, evidence):
    cv = pbn.CrossValidation(data, 10, seed)
    loglik = 0
    for train_df, test_df in cv:
        if isinstance(variable, str):
            node_data = train_df.to_pandas().loc[:, [variable] + evidence].dropna()
            variable_data = node_data.loc[:, variable]
            evidence_data = node_data.loc[:, evidence]
            test_node_data = test_df.to_pandas().loc[:, [variable] + evidence].dropna()
            test_variable_data = test_node_data.loc[:, variable]
            test_evidence_data = test_node_data.loc[:, evidence]
        else:
            node_data = train_df.to_pandas().iloc[:, [variable] + evidence].dropna()
            variable_data = node_data.iloc[:, 0]
            evidence_data = node_data.iloc[:, 1:]
            test_node_data = test_df.to_pandas().iloc[:, [variable] + evidence].dropna()
            test_variable_data = test_node_data.iloc[:, 0]
            test_evidence_data = test_node_data.iloc[:, 1:]

        if node_type == pbn.LinearGaussianCPDType():
            N = variable_data.shape[0]
            d = evidence_data.shape[1]
            linregress_data = np.column_stack((np.ones(N), evidence_data.to_numpy()))
            (beta, res, _, _) = np.linalg.lstsq(linregress_data, variable_data.to_numpy(), rcond=None)
            var = res / (N - d - 1)

            means = beta[0] + np.sum(beta[1:]*test_evidence_data, axis=1)
            loglik += norm.logpdf(test_variable_data, means, np.sqrt(var)).sum()
        elif node_type == pbn.CKDEType():
            k_joint = gaussian_kde(node_data.to_numpy().T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())
            if evidence:
                k_marg = gaussian_kde(evidence_data.to_numpy().T, bw_method=k_joint.covariance_factor())
                loglik += np.sum(k_joint.logpdf(test_node_data.to_numpy().T) - k_marg.logpdf(test_evidence_data.to_numpy().T))
            else:
                loglik += np.sum(k_joint.logpdf(test_node_data.to_numpy().T))

    return loglik

def test_cvl_create():
    s = pbn.CVLikelihood(df)
    assert len(list(s.cv)) == 10
    s = pbn.CVLikelihood(df, 5)
    assert len(list(s.cv)) == 5
    
    s = pbn.CVLikelihood(df, 10, 0)
    assert len(list(s.cv)) == 10
    s2 = pbn.CVLikelihood(df, 10, 0)
    assert len(list(s2.cv)) == 10

    for (train_cv, test_cv), (train_cv2, test_cv2) in zip(s.cv, s2.cv):
        assert train_cv.equals(train_cv2), "Train CV DataFrames with the same seed are not equal."
        assert test_cv.equals(test_cv2), "Test CV DataFrames with the same seed are not equal."

    with pytest.raises(ValueError) as ex:
        s = pbn.CVLikelihood(df, SIZE+1)
    assert "Cannot split" in str(ex.value)

def test_cvl_local_score_gbn():
    gbn = pbn.GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])
    
    cvl = pbn.CVLikelihood(df, 10, seed)
    
    assert np.isclose(cvl.local_score(gbn, 'a', []),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'a', []))
    assert np.isclose(cvl.local_score(gbn, 'b', ['a']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'b', ['a']))
    assert np.isclose(cvl.local_score(gbn, 'c', ['a', 'b']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'c', ['a', 'b']))
    assert np.isclose(cvl.local_score(gbn, 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'd', ['a', 'b', 'c']))
    assert np.isclose(cvl.local_score(gbn, 'd', ['a', 'b', 'c']),
                      cvl.local_score(gbn, 'd', ['b', 'c', 'a']))

    assert cvl.local_score(gbn, 'a') == cvl.local_score(gbn, 'a', gbn.parents('a'))
    assert cvl.local_score(gbn, 'b') == cvl.local_score(gbn, 'b', gbn.parents('b'))
    assert cvl.local_score(gbn, 'c') == cvl.local_score(gbn, 'c', gbn.parents('c'))
    assert cvl.local_score(gbn, 'd') == cvl.local_score(gbn, 'd', gbn.parents('d'))

def test_cvl_local_score_gbn_null():
    gbn = pbn.GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])
    
    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    cvl = pbn.CVLikelihood(df_null, 10, seed)

    assert np.isclose(cvl.local_score(gbn, 'a', []),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'a', []))
    assert np.isclose(cvl.local_score(gbn, 'b', ['a']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'b', ['a']))
    assert np.isclose(cvl.local_score(gbn, 'c', ['a', 'b']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'c', ['a', 'b']))
    assert np.isclose(cvl.local_score(gbn, 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'd', ['a', 'b', 'c']))
    assert np.isclose(cvl.local_score(gbn, 'd', ['a', 'b', 'c']),
                      cvl.local_score(gbn, 'd', ['b', 'c', 'a']))

    assert cvl.local_score(gbn, 'a') == cvl.local_score(gbn, 'a', gbn.parents('a'))
    assert cvl.local_score(gbn, 'b') == cvl.local_score(gbn, 'b', gbn.parents('b'))
    assert cvl.local_score(gbn, 'c') == cvl.local_score(gbn, 'c', gbn.parents('c'))
    assert cvl.local_score(gbn, 'd') == cvl.local_score(gbn, 'd', gbn.parents('d'))

def test_cvl_local_score_spbn():
    spbn = pbn.SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')],
                            [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    
    cvl = pbn.CVLikelihood(df, 10, seed)

    assert np.isclose(cvl.local_score(spbn, 'a', []),
                      numpy_local_score(pbn.CKDEType(), df, 'a', []))
    assert np.isclose(cvl.local_score(spbn, 'b', ['a']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'b', ['a']))
    assert np.isclose(cvl.local_score(spbn, 'c', ['a', 'b']),
                      numpy_local_score(pbn.CKDEType(), df, 'c', ['a', 'b']))
    assert np.isclose(cvl.local_score(spbn, 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'd', ['a', 'b', 'c']))
    assert np.isclose(cvl.local_score(spbn, 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'd', ['b', 'c', 'a']))

    assert cvl.local_score(spbn, 'a') == cvl.local_score(spbn, 'a', spbn.parents('a'))
    assert cvl.local_score(spbn, 'b') == cvl.local_score(spbn, 'b', spbn.parents('b'))
    assert cvl.local_score(spbn, 'c') == cvl.local_score(spbn, 'c', spbn.parents('c'))
    assert cvl.local_score(spbn, 'd') == cvl.local_score(spbn, 'd', spbn.parents('d'))

    assert np.isclose(cvl.local_score_node_type(spbn, pbn.LinearGaussianCPDType(), 'a', []),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'a', []))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.CKDEType(), 'b', ['a']),
                      numpy_local_score(pbn.CKDEType(), df, 'b', ['a']))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.LinearGaussianCPDType(), 'c', ['a', 'b']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df, 'c', ['a', 'b']))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.CKDEType(), 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.CKDEType(), df, 'd', ['a', 'b', 'c']))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.CKDEType(), 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.CKDEType(), df, 'd', ['b', 'c', 'a']))


def test_cvl_local_score_null_spbn():
    spbn = pbn.SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')],
                            [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    
    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    cvl = pbn.CVLikelihood(df_null, 10, seed)

    assert np.isclose(cvl.local_score(spbn, 'a', []), 
                      numpy_local_score(pbn.CKDEType(), df_null, 'a', []))
    assert np.isclose(cvl.local_score(spbn, 'b', ['a']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'b', ['a']))
    assert np.isclose(cvl.local_score(spbn, 'c', ['a', 'b']), 
                      numpy_local_score(pbn.CKDEType(), df_null, 'c', ['a', 'b']))
    assert np.isclose(cvl.local_score(spbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'd', ['a', 'b', 'c']))
    assert np.isclose(cvl.local_score(spbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'd', ['b', 'c', 'a']))

    assert cvl.local_score(spbn, 'a') == cvl.local_score(spbn, 'a', spbn.parents('a'))
    assert cvl.local_score(spbn, 'b') == cvl.local_score(spbn, 'b', spbn.parents('b'))
    assert cvl.local_score(spbn, 'c') == cvl.local_score(spbn, 'c', spbn.parents('c'))
    assert cvl.local_score(spbn, 'd') == cvl.local_score(spbn, 'd', spbn.parents('d'))

    assert np.isclose(cvl.local_score_node_type(spbn, pbn.LinearGaussianCPDType(), 'a', []),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'a', []))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.CKDEType(), 'b', ['a']),
                      numpy_local_score(pbn.CKDEType(), df_null, 'b', ['a']))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.LinearGaussianCPDType(), 'c', ['a', 'b']),
                      numpy_local_score(pbn.LinearGaussianCPDType(), df_null, 'c', ['a', 'b']))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.CKDEType(), 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.CKDEType(), df_null, 'd', ['a', 'b', 'c']))
    assert np.isclose(cvl.local_score_node_type(spbn, pbn.CKDEType(), 'd', ['a', 'b', 'c']),
                      numpy_local_score(pbn.CKDEType(), df_null, 'd', ['b', 'c', 'a']))

def test_cvl_score():
    gbn = pbn.GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    cv = pbn.CVLikelihood(df, 10, 0)

    assert np.isclose(cv.score(gbn), (
                            cv.local_score(gbn, 'a', []) +
                            cv.local_score(gbn, 'b', ['a']) +
                            cv.local_score(gbn, 'c', ['a', 'b']) +
                            cv.local_score(gbn, 'd', ['a', 'b', 'c'])))

    spbn = pbn.SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')], 
                            [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])

    assert np.isclose(cv.score(spbn), (
                            cv.local_score(spbn, 'a') +
                            cv.local_score(spbn, 'b') +
                            cv.local_score(spbn, 'c') +
                            cv.local_score(spbn, 'd')))


import pytest
import pybnesian as pbn
import pickle
import itertools

@pytest.fixture
def gaussian_type_bytes():
    g = pbn.GaussianNetworkType()
    return pickle.dumps(g)

@pytest.fixture
def spbn_type_bytes():
    s = pbn.SemiparametricBNType()
    return pickle.dumps(s)

@pytest.fixture
def kde_type_bytes():
    k = pbn.KDENetworkType()
    return pickle.dumps(k)

@pytest.fixture
def discrete_type_bytes():
    d = pbn.DiscreteBNType()
    return pickle.dumps(d)

class NewBNType(pbn.BayesianNetworkType):
    def __init__(self):
        pbn.BayesianNetworkType.__init__(self)

    def __str__(self):
        return "NewType"

@pytest.fixture
def new_type_bytes():
    nn = NewBNType()
    return pickle.dumps(nn)

class OtherBNType(pbn.BayesianNetworkType):
    def __init__(self):
        pbn.BayesianNetworkType.__init__(self)
        self.some_useful_info = "info"

    def __str__(self):
        return "OtherType"

    def __getstate_extra__(self):
        return self.some_useful_info
    
    def __setstate_extra__(self, extra):
        self.some_useful_info = extra

@pytest.fixture
def other_type_bytes():
    o = OtherBNType()
    return pickle.dumps(o)


def test_serialization_bn_type(gaussian_type_bytes, spbn_type_bytes, kde_type_bytes,
                               discrete_type_bytes, new_type_bytes, other_type_bytes):
    loaded_g = pickle.loads(gaussian_type_bytes)
    new_g = pbn.GaussianNetworkType()
    assert loaded_g == new_g

    loaded_s = pickle.loads(spbn_type_bytes)
    new_s = pbn.SemiparametricBNType()
    assert loaded_s == new_s

    loaded_k = pickle.loads(kde_type_bytes)
    new_k = pbn.KDENetworkType()
    assert loaded_k == new_k

    loaded_d = pickle.loads(discrete_type_bytes)
    new_d = pbn.DiscreteBNType()
    assert loaded_d == new_d

    loaded_nn = pickle.loads(new_type_bytes)
    new_nn = NewBNType()
    assert loaded_nn == new_nn

    loaded_o = pickle.loads(other_type_bytes)
    new_o = OtherBNType()
    assert loaded_o == new_o
    assert loaded_o.some_useful_info == "info"

    m = [loaded_g, loaded_s, loaded_k, loaded_d, loaded_nn, loaded_o]

    for t in itertools.combinations(m, 2):
        assert t[0] != t[1]


import pytest
import numpy as np
import pyarrow as pa
import pybnesian as pbn
from pybnesian import BandwidthSelector
from scipy.stats import gaussian_kde

import util_test

SIZE = 500
df = util_test.generate_normal_data(SIZE, seed=0)
df_float = df.astype('float32')

def test_check_type():
    cpd = pbn.KDE(['a'])
    cpd.fit(df)
    with pytest.raises(ValueError) as ex:
        cpd.logl(df_float)
    assert "Data type of training and test datasets is different." in str(ex.value)
    with pytest.raises(ValueError) as ex:
        cpd.slogl(df_float)
    assert "Data type of training and test datasets is different." in str(ex.value)

    cpd.fit(df_float)
    with pytest.raises(ValueError) as ex:
        cpd.logl(df)
    assert "Data type of training and test datasets is different." in str(ex.value)
    with pytest.raises(ValueError) as ex:
        cpd.slogl(df)
    assert "Data type of training and test datasets is different." in str(ex.value)

def test_kde_variables():
    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        cpd = pbn.KDE(variables)
        assert cpd.variables() == variables

def test_kde_bandwidth():
    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        for instances in [50, 1000, 10000]:
            npdata = df.loc[:, variables].to_numpy()
            # Test normal reference rule
            scipy_kde = gaussian_kde(npdata[:instances, :].T,
                            bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

            cpd = pbn.KDE(variables)
            cpd.fit(df.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, scipy_kde.covariance)), "Wrong bandwidth computed with normal reference rule."

            cpd.fit(df_float.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, scipy_kde.covariance)), "Wrong bandwidth computed with normal reference rule."

            scipy_kde = gaussian_kde(npdata[:instances, :].T)

            cpd = pbn.KDE(variables, pbn.ScottsBandwidth())
            cpd.fit(df.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, scipy_kde.covariance)), "Wrong bandwidth computed with Scott's rule."

            cpd.fit(df_float.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, scipy_kde.covariance)), "Wrong bandwidth computed with Scott's rule."


    cpd = pbn.KDE(['a'])
    cpd.fit(df)
    cpd.bandwidth = [[1]]
    assert cpd.bandwidth == np.asarray([[1]]), "Could not change bandwidth."

    cpd.fit(df_float)
    cpd.bandwidth = [[1]]
    assert cpd.bandwidth == np.asarray([[1]]), "Could not change bandwidth."

class UnitaryBandwidth(BandwidthSelector):
    def __init__(self):
        BandwidthSelector.__init__(self)

    def bandwidth(self, df, variables):
        return np.eye(len(variables))

def test_kde_new_bandwidth():
    kde = pbn.KDE(["a"], UnitaryBandwidth())
    kde.fit(df)
    assert kde.bandwidth == np.eye(1)

    kde.fit(df_float)
    assert kde.bandwidth == np.eye(1)

    kde = pbn.KDE(["a", "b", "c", "d"], UnitaryBandwidth())
    kde.fit(df)
    assert np.all(kde.bandwidth == np.eye(4))

    kde.fit(df_float)
    assert np.all(kde.bandwidth == np.eye(4))

def test_kde_data_type():
    k = pbn.KDE(["a"])

    with pytest.raises(ValueError) as ex:
        k.data_type()
    "KDE factor not fitted" in str(ex.value)

    k.fit(df)
    assert k.data_type() == pa.float64()
    k.fit(df_float)
    assert k.data_type() == pa.float32()


def test_kde_fit():
    def _test_kde_fit_iter(variables, _df, instances):
        cpd = pbn.KDE(variables)
        assert not cpd.fitted()
        cpd.fit(_df.iloc[:instances,:])
        assert cpd.fitted()

        npdata = _df.loc[:, variables].to_numpy()
        scipy_kde = gaussian_kde(npdata[:instances, :].T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        assert scipy_kde.n == cpd.num_instances(), "Wrong number of training instances."
        assert scipy_kde.d == cpd.num_variables(), "Wrong number of training variables."

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        for instances in [50, 150, 500]:
            _test_kde_fit_iter(variables, df, instances)
            _test_kde_fit_iter(variables, df_float, instances)

def test_kde_fit_null():
    def _test_kde_fit_null_iter(variables, _df, instances):
        cpd = pbn.KDE(variables)
        assert not cpd.fitted()
        cpd.fit(_df.iloc[:instances,:])
        assert cpd.fitted()

        npdata = _df.loc[:, variables].to_numpy()
        npdata_instances = npdata[:instances,:]

        nan_rows = np.any(np.isnan(npdata_instances), axis=1)
        npdata_no_null = npdata_instances[~nan_rows,:]
        scipy_kde = gaussian_kde(npdata_no_null.T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        assert scipy_kde.n == cpd.num_instances(), "Wrong number of training instances with null values."
        assert scipy_kde.d == cpd.num_variables(), "Wrong number of training variables with null values."
        assert np.all(np.isclose(scipy_kde.covariance, cpd.bandwidth)), "Wrong bandwidth with null values."

    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        for instances in [50, 150, 500]:
            _test_kde_fit_null_iter(variables, df_null, instances)
            _test_kde_fit_null_iter(variables, df_null_float, instances)

def test_kde_logl():
    def _test_kde_logl_iter(variables, _df, _test_df):
        cpd = pbn.KDE(variables)
        cpd.fit(_df)

        npdata = _df.loc[:, variables].to_numpy()
        scipy_kde = gaussian_kde(npdata.T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        test_npdata = _test_df.loc[:, variables].to_numpy()

        logl = cpd.logl(_test_df)
        scipy = scipy_kde.logpdf(test_npdata.T)

        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(logl, scipy, atol=0.0005))
        else:
            assert np.all(np.isclose(logl, scipy))

    test_df = util_test.generate_normal_data(50, seed=1)
    test_df_float = test_df.astype('float32')

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_kde_logl_iter(variables, df, test_df)
        _test_kde_logl_iter(variables, df_float, test_df_float)

    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.logl(test_df), cpd2.logl(test_df))), "Order of evidence changes logl() result."

    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.logl(test_df_float), cpd2.logl(test_df_float))), "Order of evidence changes logl() result."

def test_kde_logl_null():
    def _test_kde_logl_null_iter(variables, _df, _test_df):
        cpd = pbn.KDE(variables)
        cpd.fit(_df)

        npdata = _df.loc[:, variables].to_numpy()
        scipy_kde = gaussian_kde(npdata.T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        test_npdata = _test_df.loc[:, variables].to_numpy()

        logl = cpd.logl(_test_df)

        scipy_result = np.full((test_npdata.shape[0],), np.nan)
        nan_rows = np.any(np.isnan(test_npdata), axis=1)
        scipy_result[~nan_rows] = scipy_kde.logpdf(test_npdata[~nan_rows].T)

        if npdata.dtype == "float32":
            assert np.all(np.isclose(logl, scipy_result, atol=0.0005, equal_nan=True))
        else:
            assert np.all(np.isclose(logl, scipy_result, equal_nan=True))

    TEST_SIZE = 50

    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    np.random.seed(0)
    a_null = np.random.randint(0, TEST_SIZE, size=10)
    b_null = np.random.randint(0, TEST_SIZE, size=10)
    c_null = np.random.randint(0, TEST_SIZE, size=10)
    d_null = np.random.randint(0, TEST_SIZE, size=10)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = test_df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_kde_logl_null_iter(variables, df, df_null)
        _test_kde_logl_null_iter(variables, df_float, df_null_float)


    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.logl(df_null), cpd2.logl(df_null), equal_nan=True)), "Order of evidence changes logl() result."

    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.logl(df_null_float), cpd2.logl(df_null_float), atol=0.0005, equal_nan=True)), "Order of evidence changes logl() result."

def test_kde_slogl():
    def _test_kde_slogl_iter(variables, _df, _test_df):
        cpd = pbn.KDE(variables)
        cpd.fit(_df)

        npdata = _df.loc[:, variables].to_numpy()
        scipy_kde = gaussian_kde(npdata.T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        test_npdata = _test_df.loc[:, variables].to_numpy()
        assert np.all(np.isclose(cpd.slogl(_test_df), scipy_kde.logpdf(test_npdata.T).sum()))


    test_df = util_test.generate_normal_data(50, seed=1)
    test_df_float = test_df.astype('float32')

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_kde_slogl_iter(variables, df, test_df)
        _test_kde_slogl_iter(variables, df_float, test_df_float)

    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.slogl(test_df), cpd2.slogl(test_df))), "Order of evidence changes slogl() result."

    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.slogl(test_df_float), cpd2.slogl(test_df_float))), "Order of evidence changes slogl() result."


def test_kde_slogl_null():
    def _test_kde_slogl_null_iter(variables, _df, _test_df):
        cpd = pbn.KDE(variables)
        cpd.fit(_df)

        npdata = _df.loc[:, variables].to_numpy()
        scipy_kde = gaussian_kde(npdata.T,
                        bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())

        test_npdata = _test_df.loc[:, variables].to_numpy()

        nan_rows = np.any(np.isnan(test_npdata), axis=1)

        assert np.all(np.isclose(cpd.slogl(_test_df), scipy_kde.logpdf(test_npdata[~nan_rows].T).sum()))

    TEST_SIZE = 50

    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    np.random.seed(0)
    a_null = np.random.randint(0, TEST_SIZE, size=10)
    b_null = np.random.randint(0, TEST_SIZE, size=10)
    c_null = np.random.randint(0, TEST_SIZE, size=10)
    d_null = np.random.randint(0, TEST_SIZE, size=10)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = test_df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_kde_slogl_null_iter(variables, df, df_null)
        _test_kde_slogl_null_iter(variables, df_float, df_null_float)


    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.slogl(df_null), cpd2.slogl(df_null))), "Order of evidence changes slogl() result."

    cpd = pbn.KDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.KDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.slogl(df_null_float), cpd2.slogl(df_null_float))), "Order of evidence changes slogl() result."



import pytest
import numpy as np
import pybnesian as pbn
import util_test

SIZE = 10000
df = util_test.generate_normal_data(SIZE)


def numpy_fit_mle_lg(data, variable, evidence):
    if isinstance(variable, str):
        node_data = data.loc[:, [variable] + evidence].dropna()
        variable_data = node_data.loc[:, variable]
        evidence_data = node_data.loc[:, evidence]
    else:
        node_data = data.iloc[:, [variable] + evidence].dropna()
        variable_data = node_data.iloc[:, 0]
        evidence_data = node_data.iloc[:, 1:]

    N = variable_data.shape[0]
    d = evidence_data.shape[1]
    linregress_data = np.column_stack((np.ones(N), evidence_data.to_numpy()))
    (beta, res, _, _) = np.linalg.lstsq(linregress_data, variable_data.to_numpy(), rcond=None)
    var = res / (N - d - 1)

    return beta, var

def test_mle_create():
    with pytest.raises(ValueError) as ex:
        mle = pbn.MLE(pbn.CKDEType())
    assert "MLE not available" in str(ex.value)

    mle = pbn.MLE(pbn.LinearGaussianCPDType())

def test_mle_lg():
    mle = pbn.MLE(pbn.LinearGaussianCPDType())

    p = mle.estimate(df, "a", [])
    np_beta, np_var = numpy_fit_mle_lg(df, "a", [])
    assert np.all(np.isclose(p.beta, np_beta))
    assert np.isclose(p.variance, np_var)

    p = mle.estimate(df, "b", ["a"])
    np_beta, np_var = numpy_fit_mle_lg(df, "b", ["a"])
    assert np.all(np.isclose(p.beta, np_beta))
    assert np.isclose(p.variance, np_var)

    p = mle.estimate(df, "c", ["a", "b"])
    np_beta, np_var = numpy_fit_mle_lg(df, "c", ["a", "b"])
    assert np.all(np.isclose(p.beta, np_beta))
    assert np.isclose(p.variance, np_var)

    p = mle.estimate(df, "d", ["a", "b", "c"])
    np_beta, np_var = numpy_fit_mle_lg(df, "d", ["a", "b", "c"])
    assert np.all(np.isclose(p.beta, np_beta))
    assert np.isclose(p.variance, np_var)


import numpy as np
import pandas as pd
import pytest
import pybnesian as pbn
from pybnesian import FactorType, Factor, LinearGaussianCPD, CKDE, DiscreteFactor
import pickle

@pytest.fixture
def lg_bytes():
    lg = LinearGaussianCPD("c", ["a", "b"])
    return pickle.dumps(lg)

@pytest.fixture
def ckde_bytes():
    ckde = CKDE("c", ["a", "b"])
    return pickle.dumps(ckde)

@pytest.fixture
def discrete_bytes():
    discrete = DiscreteFactor("c", ["a", "b"])
    return pickle.dumps(discrete)

class NewType(FactorType):
    def __init__(self, factor_class):
        FactorType.__init__(self)
        self.factor_class = factor_class

    def new_factor(self, model, variable, evidence):
        return self.factor_class(variable, evidence)

    def __str__(self):
        return "NewType"

class NewFactor(Factor):
    def __init__(self, variable, evidence):
        Factor.__init__(self, variable, evidence)
        self._fitted = False
        self.some_fit_data = None
    
    def fit(self, df):
        self.some_fit_data = "fitted"
        self._fitted = True
    
    def fitted(self):
        return self._fitted

    def type(self):
        return NewType(NewFactor)

    def __str__(self):
        return "NewFactor"

    def __getstate_extra__(self):
        d = {'fitted': self._fitted,  'some_fit_data': self.some_fit_data}
        return d

    def __setstate_extra__(self, d):
        self._fitted = d['fitted']
        self.some_fit_data = d['some_fit_data']

class NewFactorBis(Factor):
    def __init__(self, variable, evidence):
        Factor.__init__(self, variable, evidence)
        self._fitted = False
        self.some_fit_data = None

    def fit(self, df):
        self.some_fit_data = "fitted"
        self._fitted = True
    
    def fitted(self):
        return self._fitted

    def type(self):
        return NewType(NewFactorBis)

    def __str__(self):
        return "NewFactor"

    def __getstate__(self):
        d = {'variable': self.variable(),
             'evidence': self.evidence(),
             'fitted': self._fitted,
             'some_fit_data': self.some_fit_data}
        return d

    def __setstate__(self, d):
        Factor.__init__(self, d['variable'], d['evidence'])
        self._fitted = d['fitted']
        self.some_fit_data = d['some_fit_data']

@pytest.fixture
def new_bytes():
    n = NewFactor("c", ["a", "b"])
    return pickle.dumps(n)

@pytest.fixture
def newbis_bytes():
    n = NewFactorBis("c", ["a", "b"])
    return pickle.dumps(n)

def test_serialization_unfitted_factor(lg_bytes, ckde_bytes, discrete_bytes, new_bytes, newbis_bytes):
    loaded_lg = pickle.loads(lg_bytes)
    assert loaded_lg.variable() == "c"
    assert set(loaded_lg.evidence()) == set(["a", "b"])
    assert not loaded_lg.fitted()
    assert loaded_lg.type() == pbn.LinearGaussianCPDType()

    loaded_ckde = pickle.loads(ckde_bytes)
    assert loaded_ckde.variable() == "c"
    assert set(loaded_ckde.evidence()) == set(["a", "b"])
    assert not loaded_ckde.fitted()
    assert loaded_ckde.type() == pbn.CKDEType()

    loaded_discrete = pickle.loads(discrete_bytes)
    assert loaded_discrete.variable() == "c"
    assert set(loaded_discrete.evidence()) == set(["a", "b"])
    assert not loaded_discrete.fitted()
    assert loaded_discrete.type() == pbn.DiscreteFactorType()

    loaded_new = pickle.loads(new_bytes)
    assert loaded_new.variable() == "c"
    assert set(loaded_new.evidence()) == set(["a", "b"])
    assert not loaded_new.fitted()
    assert type(loaded_new.type()) == NewType
    nn = NewFactor("a", [])
    assert loaded_new.type() == nn.type()

    from pybnesian import GaussianNetwork
    dummy_network = GaussianNetwork(["a", "b", "c", "d"])
    assert type(loaded_new.type().new_factor(dummy_network, "a", [])) == NewFactor

    loaded_newbis = pickle.loads(newbis_bytes)
    assert loaded_newbis.variable() == "c"
    assert set(loaded_newbis.evidence()) == set(["a", "b"])
    assert not loaded_newbis.fitted()
    assert type(loaded_newbis.type()) == NewType
    nnbis = NewFactorBis("a", [])
    assert loaded_newbis.type() == nnbis.type()
    assert type(loaded_newbis.type().new_factor(dummy_network, "a", [])) == NewFactorBis

    assert loaded_lg.type() != loaded_ckde.type()
    assert loaded_lg.type() != loaded_discrete.type()
    assert loaded_lg.type() != loaded_new.type()
    assert loaded_ckde.type() != loaded_discrete.type()
    assert loaded_ckde.type() != loaded_new.type()
    assert loaded_discrete.type() != loaded_new.type()
    assert loaded_newbis.type() == loaded_new.type()

@pytest.fixture
def lg_fitted_bytes():
    lg = LinearGaussianCPD("c", ["a", "b"], [1, 2, 3], 0.5)
    return pickle.dumps(lg)

@pytest.fixture
def ckde_fitted_bytes():
    np.random.seed(1)
    data = pd.DataFrame({'a': np.random.rand(10), 'b': np.random.rand(10), 'c': np.random.rand(10)}).astype(float)
    ckde = CKDE("c", ["a", "b"])
    ckde.fit(data)
    return pickle.dumps(ckde)

@pytest.fixture
def discrete_fitted_bytes():
    discrete = DiscreteFactor("c", ["a", "b"])

    data = pd.DataFrame({'a': ["a1", "a2", "a1", "a2", "a2", "a2", "a2", "a2"], 
                         'b': ["b1", "b1", "b1", "b1", "b1", "b2", "b1", "b2"],
                         'c': ["c1", "c1", "c1", "c1", "c2", "c2", "c2", "c2"]}, dtype="category")
    discrete.fit(data)
    return pickle.dumps(discrete)

@pytest.fixture
def new_fitted_bytes():
    n = NewFactor("c", ["a", "b"])
    n.fit(None)
    return pickle.dumps(n)

@pytest.fixture
def newbis_fitted_bytes():
    n = NewFactorBis("c", ["a", "b"])
    n.fit(None)
    return pickle.dumps(n)

def test_serialization_fitted_factor(lg_fitted_bytes, ckde_fitted_bytes, discrete_fitted_bytes, new_fitted_bytes,
                                     newbis_fitted_bytes):
    loaded_lg = pickle.loads(lg_fitted_bytes)
    assert loaded_lg.variable() == "c"
    assert set(loaded_lg.evidence()) == set(["a", "b"])
    assert loaded_lg.fitted()
    assert list(loaded_lg.beta) == [1, 2, 3]
    assert loaded_lg.variance == 0.5

    loaded_ckde = pickle.loads(ckde_fitted_bytes)
    assert loaded_ckde.variable() == "c"
    assert set(loaded_ckde.evidence()) == set(["a", "b"])
    assert loaded_ckde.fitted()
    assert loaded_ckde.type() == pbn.CKDEType()
    assert loaded_ckde.num_instances() == 10
    tr = loaded_ckde.kde_joint().dataset().to_pandas()
    np.random.seed(1)
    assert np.all(tr['a'] == np.random.rand(10))
    assert np.all(tr['b'] == np.random.rand(10))
    assert np.all(tr['c'] == np.random.rand(10))

    loaded_discrete = pickle.loads(discrete_fitted_bytes)
    assert loaded_discrete.variable() == "c"
    assert set(loaded_discrete.evidence()) == set(["a", "b"])
    assert loaded_discrete.fitted()
    assert loaded_discrete.type() == pbn.DiscreteFactorType()

    test = pd.DataFrame({'a': ["a1", "a2", "a1", "a2", "a1", "a2", "a1", "a2"], 
                         'b': ["b1", "b1", "b2", "b2", "b1", "b1", "b2", "b2"],
                         'c': ["c1", "c1", "c1", "c1", "c2", "c2", "c2", "c2"]}, dtype="category")
    ll = loaded_discrete.logl(test)
    assert list(np.exp(ll)) == [1, 0.5, 0.5, 0, 0, 0.5, 0.5, 1]

    loaded_new = pickle.loads(new_fitted_bytes)
    assert loaded_new.variable() == "c"
    assert set(loaded_new.evidence()) == set(["a", "b"])
    assert loaded_new.fitted()
    assert type(loaded_new.type()) == NewType
    nn = NewFactor("a", [])
    assert loaded_new.type() == nn.type()
    assert loaded_new.some_fit_data == "fitted"

    loaded_newbis = pickle.loads(newbis_fitted_bytes)
    assert loaded_newbis.variable() == "c"
    assert set(loaded_newbis.evidence()) == set(["a", "b"])
    assert loaded_newbis.fitted()
    assert type(loaded_newbis.type()) == NewType
    nn = NewFactorBis("a", [])
    assert loaded_newbis.type() == nn.type()
    assert loaded_newbis.some_fit_data == "fitted"
    assert type(loaded_newbis.type()) == type(loaded_new.type())

    assert loaded_lg.type() != loaded_ckde.type()
    assert loaded_lg.type() != loaded_discrete.type()
    assert loaded_lg.type() != loaded_new.type()
    assert loaded_ckde.type() != loaded_discrete.type()
    assert loaded_ckde.type() != loaded_new.type()
    assert loaded_discrete.type() != loaded_new.type()


import numpy as np
import pandas as pd
import pybnesian as pbn

import util_test

SIZE = 10000

df = util_test.generate_normal_data(SIZE)

def test_holdout_disjoint():
    hold = pbn.HoldOut(df)

    train_df, test_df = hold.training_data(), hold.test_data()

    assert (train_df.num_rows + test_df.num_rows) == SIZE, "HoldOut do not have the expected number of rows"

    assert train_df.num_rows == round((1-0.2) * df.shape[0]), "Train DataFrame do not have the expected number of instances"
    assert test_df.num_rows == round(0.2 * df.shape[0]), "Test DataFrame do not have the expected number of instances"

    combination = pd.concat([train_df.to_pandas(), test_df.to_pandas()])

    assert df.sort_values("a", axis=0).reset_index(drop=True)\
                .equals(combination.sort_values("a", axis=0).reset_index(drop=True)),\
                 "The combination of train and test dataset is not equal to the original DataFrame."
    
    hold = pbn.HoldOut(df, test_ratio=0.3)
    train_df, test_df = hold.training_data(), hold.test_data()

    assert (train_df.num_rows + test_df.num_rows) == SIZE, "HoldOut do not have the expected number of rows"

    assert train_df.num_rows == round((1-0.3) * df.shape[0]), "Train DataFrame do not have the expected number of instances"
    assert test_df.num_rows == round(0.3 * df.shape[0]), "Test DataFrame do not have the expected number of instances"

    combination = pd.concat([train_df.to_pandas(), test_df.to_pandas()])

    assert df.sort_values("a", axis=0).reset_index(drop=True)\
                .equals(combination.sort_values("a", axis=0).reset_index(drop=True)),\
                 "The combination of train and test dataset is not equal to the original DataFrame."

def test_holdout_seed():
    hold = pbn.HoldOut(df, seed=0)
    hold2 = pbn.HoldOut(df, seed=0)

    train_df, test_df = hold.training_data(), hold.test_data()
    train_df2, test_df2 = hold2.training_data(), hold2.test_data()

    assert train_df.equals(train_df2), "Train CV DataFrames with the same seed are not equal."
    assert test_df.equals(test_df2), "Test CV DataFrames with the same seed are not equal."

    hold3 = pbn.HoldOut(df, seed=1)
    train_df3, test_df3 = hold3.training_data(), hold3.test_data()

    assert not train_df.equals(train_df3), "Train CV DataFrames with different seeds return the same result."
    assert not test_df.equals(test_df3), "Test CV DataFrames with different seeds return the same result."

def test_holdout_null():
    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    non_null = df_null.dropna()
    hold = pbn.HoldOut(df_null)

    train_df, test_df = hold.training_data(), hold.test_data()

    assert (train_df.num_rows + test_df.num_rows) == non_null.shape[0], "HoldOut do not have the expected number of rows"
    assert train_df.num_rows == round((1-0.2) * non_null.shape[0]), "Train DataFrame do not have the expected number of instances"
    assert test_df.num_rows == round(0.2 * non_null.shape[0]), "Test DataFrame do not have the expected number of instances"

    combination = pd.concat([train_df.to_pandas(), test_df.to_pandas()])

    assert combination.sort_values("a", axis=0).reset_index(drop=True)\
                .equals(non_null.sort_values("a", axis=0).reset_index(drop=True)),\
                 "The combination of train and test dataset is not equal to the original DataFrame."

    hold_null = pbn.HoldOut(df_null, include_null=True)
    train_df, test_df = hold_null.training_data(), hold_null.test_data()
    assert (train_df.num_rows + test_df.num_rows) == SIZE, "HoldOut do not have the expected number of rows"
    assert train_df.num_rows == round((1-0.2) * SIZE), "Train DataFrame do not have the expected number of instances"
    assert test_df.num_rows == round(0.2 * SIZE), "Test DataFrame do not have the expected number of instances"

    combination = pd.concat([train_df.to_pandas(), test_df.to_pandas()])

    assert combination.sort_values(["a", "b", "c", "d"], axis=0).reset_index(drop=True)\
                .equals(df.sort_values(["a", "b", "c", "d"], axis=0).reset_index(drop=True)),\
                 "The combination of train and test dataset is not equal to the original DataFrame."

    


import pytest
import re
import numpy as np
import pandas as pd
from scipy.stats import norm
import pybnesian as pbn
from pybnesian import GaussianNetwork, ConditionalGaussianNetwork, DynamicGaussianNetwork
import util_test

df = util_test.generate_normal_data(1000)

def test_create_dbn():
    variables = ["a", "b", "c", "d"]
    gbn = DynamicGaussianNetwork(variables, 2)

    assert gbn.markovian_order() == 2
    assert gbn.variables() == ["a", "b", "c", "d"]
    assert gbn.num_variables() == 4
    assert gbn.type() == pbn.GaussianNetworkType()

    transition_nodes = [v + "_t_0" for v in variables]
    static_nodes = [v + "_t_" + str(m) for v in variables for m in range(1, 3)]

    assert set(gbn.static_bn().nodes()) == set(static_nodes)
    assert set(gbn.transition_bn().interface_nodes()) == set(static_nodes)
    assert set(gbn.transition_bn().nodes()) == set(transition_nodes)

    static_bn = GaussianNetwork(static_nodes)
    transition_bn = ConditionalGaussianNetwork(transition_nodes, static_nodes)

    gbn2 = DynamicGaussianNetwork(variables, 2, static_bn, transition_bn)

    wrong_transition_bn = pbn.ConditionalDiscreteBN(transition_nodes, static_nodes)

    with pytest.raises(ValueError) as ex:
        gbn3 = DynamicGaussianNetwork(variables, 2, static_bn, wrong_transition_bn)
    assert "Static and transition Bayesian networks do not have the same type" in str(ex.value)

    wrong_static_bn = pbn.DiscreteBN(static_nodes)
    with pytest.raises(ValueError) as ex:
        gbn4 = DynamicGaussianNetwork(variables, 2, wrong_static_bn, wrong_transition_bn)
    assert "Bayesian networks are not Gaussian." in str(ex.value)

def test_variable_operations_dbn():
    variables = ["a", "b", "c", "d"]
    gbn = DynamicGaussianNetwork(variables, 2)

    assert gbn.markovian_order() == 2
    assert gbn.variables() == ["a", "b", "c", "d"]
    assert gbn.num_variables() == 4

    assert gbn.contains_variable("a")
    assert gbn.contains_variable("b")
    assert gbn.contains_variable("c")
    assert gbn.contains_variable("d")

    gbn.add_variable("e")
    assert set(gbn.variables()) == set(["a", "b", "c", "d", "e"])
    assert gbn.num_variables() == 5

    assert set(gbn.static_bn().nodes()) == set([v + "_t_" + str(m) for v in variables + ["e"] for m in range(1, 3)])
    assert set(gbn.transition_bn().nodes()) == set([v + "_t_0" for v in variables + ["e"]])

    gbn.remove_variable("b")
    assert set(gbn.variables()) == set(["a", "c", "d", "e"])
    assert gbn.num_variables() == 4
    assert set(gbn.static_bn().nodes()) == set([v + "_t_" + str(m) for v in ["a", "c", "d", "e"] for m in range(1, 3)])
    assert set(gbn.transition_bn().nodes()) == set([v + "_t_0" for v in ["a", "c", "d", "e"]])


def test_fit_dbn():
    variables = ["a", "b", "c", "d"]
    gbn = DynamicGaussianNetwork(variables, 2)
    assert not gbn.fitted()
    assert not gbn.static_bn().fitted()
    assert not gbn.transition_bn().fitted()
    gbn.fit(df)
    assert gbn.fitted()

    ddf = pbn.DynamicDataFrame(df, 2)
    gbn2 = DynamicGaussianNetwork(variables, 2)
    gbn2.static_bn().fit(ddf.static_df())
    assert not gbn2.fitted()
    assert gbn2.static_bn().fitted()
    assert not gbn2.transition_bn().fitted()

    gbn2.transition_bn().fit(ddf.transition_df())
    assert gbn2.fitted()
    assert gbn2.static_bn().fitted()
    assert gbn2.transition_bn().fitted()

def lg_logl_row(row, variable, evidence, beta, variance):
    m = beta[0] + beta[1:].dot(row[evidence])
    return norm(m, np.sqrt(variance)).logpdf(row[variable])

def static_logl(dbn, test_data, index, variable):
    sl = test_data.head(dbn.markovian_order())

    node_name = variable + "_t_" + str(dbn.markovian_order() - index)
    cpd = dbn.static_bn().cpd(node_name)
    evidence = cpd.evidence()

    row_values = [sl.loc[index, variable]]
    for e in evidence:
        m = re.search('(.*)_t_(\\d+)', e)
        e_var = m[1]
        t = int(m[2])

        row_values.append(sl.loc[dbn.markovian_order()-t, e_var])

    r = pd.Series(data=row_values, index=[node_name] + evidence)

    return lg_logl_row(r, node_name, evidence, cpd.beta, cpd.variance)

def transition_logl(dbn, test_data, index, variable):
    node_name = variable + "_t_0"
    cpd = dbn.transition_bn().cpd(node_name)
    evidence = cpd.evidence()

    row_values = [test_data.loc[index, variable]]
    for e in evidence:
        m = re.search('(.*)_t_(\\d+)', e)
        e_var = m[1]
        t = int(m[2])

        row_values.append(test_data.loc[index-t, e_var])

    r = pd.Series(data=row_values, index=[node_name] + evidence)
    return lg_logl_row(r, node_name, evidence, cpd.beta, cpd.variance)


def numpy_logl(dbn, test_data):
    ll = np.zeros((test_data.shape[0],))

    for i in range(dbn.markovian_order()):
        for v in dbn.variables():
            ll[i] += static_logl(dbn, test_data, i, v)

    for i in range(dbn.markovian_order(), test_data.shape[0]):
        for v in dbn.variables():
            ll[i] += transition_logl(dbn, test_data, i, v)

    return ll

def test_logl_dbn():
    variables = ["a", "b", "c", "d"]

    static_bn = GaussianNetwork(["a", "b", "c", "d"], [("a", "c"), ("b", "c"), ("c", "d")])
    static_bn = GaussianNetwork(["a", "b", "c", "d"], [("a", "c"), ("b", "c"), ("c", "d")])
    gbn = DynamicGaussianNetwork(variables, 2)

    static_bn = gbn.static_bn()
    static_bn.add_arc("a_t_2", "c_t_2")
    static_bn.add_arc("b_t_2", "c_t_2")
    static_bn.add_arc("c_t_2", "d_t_2")
    static_bn.add_arc("a_t_1", "c_t_1")
    static_bn.add_arc("b_t_1", "c_t_1")
    static_bn.add_arc("c_t_1", "d_t_1")

    transition_bn = gbn.transition_bn()
    transition_bn.add_arc("a_t_2", "a_t_0")
    transition_bn.add_arc("b_t_2", "b_t_0")
    transition_bn.add_arc("c_t_2", "c_t_0")
    transition_bn.add_arc("d_t_2", "d_t_0")
    transition_bn.add_arc("a_t_1", "a_t_0")
    transition_bn.add_arc("b_t_1", "b_t_0")
    transition_bn.add_arc("c_t_1", "c_t_0")
    transition_bn.add_arc("d_t_1", "d_t_0")

    gbn.fit(df)

    test_df = util_test.generate_normal_data(100)
    ground_truth_ll = numpy_logl(gbn, util_test.generate_normal_data(100))
    ll = gbn.logl(test_df)
    assert np.all(np.isclose(ground_truth_ll, ll))

def test_slogl_dbn():
    variables = ["a", "b", "c", "d"]

    static_bn = GaussianNetwork(["a", "b", "c", "d"], [("a", "c"), ("b", "c"), ("c", "d")])
    static_bn = GaussianNetwork(["a", "b", "c", "d"], [("a", "c"), ("b", "c"), ("c", "d")])
    gbn = DynamicGaussianNetwork(variables, 2)

    static_bn = gbn.static_bn()
    static_bn.add_arc("a_t_2", "c_t_2")
    static_bn.add_arc("b_t_2", "c_t_2")
    static_bn.add_arc("c_t_2", "d_t_2")
    static_bn.add_arc("a_t_1", "c_t_1")
    static_bn.add_arc("b_t_1", "c_t_1")
    static_bn.add_arc("c_t_1", "d_t_1")

    transition_bn = gbn.transition_bn()
    transition_bn.add_arc("a_t_2", "a_t_0")
    transition_bn.add_arc("b_t_2", "b_t_0")
    transition_bn.add_arc("c_t_2", "c_t_0")
    transition_bn.add_arc("d_t_2", "d_t_0")
    transition_bn.add_arc("a_t_1", "a_t_0")
    transition_bn.add_arc("b_t_1", "b_t_0")
    transition_bn.add_arc("c_t_1", "c_t_0")
    transition_bn.add_arc("d_t_1", "d_t_0")

    gbn.fit(df)
    test_df = util_test.generate_normal_data(100)
    ll = numpy_logl(gbn, test_df)
    assert np.isclose(gbn.slogl(test_df), ll.sum())


import pytest
import numpy as np
import pybnesian as pbn
from pybnesian import SemiparametricBN, LinearGaussianCPD, CKDE
import util_test

df = util_test.generate_normal_data(10000)

def test_create_spbn():
    spbn = SemiparametricBN(['a', 'b', 'c', 'd'])
    assert spbn.num_nodes() == 4
    assert spbn.num_arcs() == 0
    assert spbn.nodes() == ['a', 'b', 'c', 'd']

    for n in spbn.nodes():
        assert spbn.node_type(n) == pbn.UnknownFactorType()

    spbn = SemiparametricBN(['a', 'b', 'c', 'd'], [('a', 'c')])
    assert spbn.num_nodes() == 4
    assert spbn.num_arcs() == 1
    assert spbn.nodes() == ['a', 'b', 'c', 'd']

    for n in spbn.nodes():
        assert spbn.node_type(n) == pbn.UnknownFactorType()

    spbn = SemiparametricBN([('a', 'c'), ('b', 'd'), ('c', 'd')])
    assert spbn.num_nodes() == 4
    assert spbn.num_arcs() == 3
    assert spbn.nodes() == ['a', 'c', 'b', 'd']

    for n in spbn.nodes():
        assert spbn.node_type(n) == pbn.UnknownFactorType()

    with pytest.raises(TypeError) as ex:
        spbn = SemiparametricBN(['a', 'b', 'c'], [('a', 'c', 'b')])
    assert "incompatible constructor arguments" in str(ex.value)
    
    with pytest.raises(IndexError) as ex:
        spbn = SemiparametricBN(['a', 'b', 'c'], [('a', 'd')])
    assert "not present in the graph" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        spbn = SemiparametricBN([('a', 'b'), ('b', 'c'), ('c', 'a')])
    assert "must be a DAG" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        spbn = SemiparametricBN(['a', 'b', 'c', 'd'], [('a', 'b'), ('b', 'c'), ('c', 'a')])
    assert "must be a DAG" in str(ex.value)


    expected_node_type = {'a': pbn.CKDEType(),
                          'b': pbn.UnknownFactorType(),
                          'c': pbn.CKDEType(),
                          'd': pbn.UnknownFactorType()}

    spbn = SemiparametricBN(['a', 'b', 'c', 'd'], [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    assert spbn.num_nodes() == 4
    assert spbn.num_arcs() == 0
    assert spbn.nodes() == ['a', 'b', 'c', 'd']

    for n in spbn.nodes():
        assert spbn.node_type(n) == expected_node_type[n]

    spbn = SemiparametricBN(['a', 'b', 'c', 'd'], [('a', 'c')], [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    assert spbn.num_nodes() == 4
    assert spbn.num_arcs() == 1
    assert spbn.nodes() == ['a', 'b', 'c', 'd']

    for n in spbn.nodes():
        assert spbn.node_type(n) == expected_node_type[n]

    spbn = SemiparametricBN([('a', 'c'), ('b', 'd'), ('c', 'd')], [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    assert spbn.num_nodes() == 4
    assert spbn.num_arcs() == 3
    assert spbn.nodes() == ['a', 'c', 'b', 'd']

    for n in spbn.nodes():
        assert spbn.node_type(n) == expected_node_type[n]

    with pytest.raises(TypeError) as ex:
        spbn = SemiparametricBN(['a', 'b', 'c'], [('a', 'c', 'b')], [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    assert "incompatible constructor arguments" in str(ex.value)
    
    with pytest.raises(IndexError) as ex:
        spbn = SemiparametricBN(['a', 'b', 'c'], [('a', 'd')], [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    assert "not present in the graph" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        spbn = SemiparametricBN([('a', 'b'), ('b', 'c'), ('c', 'a')], [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    assert "must be a DAG" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        spbn = SemiparametricBN(['a', 'b', 'c', 'd'],
                                [('a', 'b'), ('b', 'c'), ('c', 'a')],
                                [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    assert "must be a DAG" in str(ex.value)


def test_node_type():
    spbn = SemiparametricBN(['a', 'b', 'c', 'd'])
    assert spbn.num_nodes() == 4
    assert spbn.num_arcs() == 0
    assert spbn.nodes() == ['a', 'b', 'c', 'd']

    for n in spbn.nodes():
        assert spbn.node_type(n) == pbn.UnknownFactorType()
    
    spbn.set_node_type('b', pbn.CKDEType())
    assert spbn.node_type('b') == pbn.CKDEType()
    spbn.set_node_type('b', pbn.LinearGaussianCPDType())
    assert spbn.node_type('b') == pbn.LinearGaussianCPDType()

def test_fit():
    spbn = SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    with pytest.raises(ValueError) as ex:
        for n in spbn.nodes():
            cpd = spbn.cpd(n)
    assert "not added" in str(ex.value)

    spbn.fit(df)

    for n in spbn.nodes():
        cpd = spbn.cpd(n)
        assert cpd.type() == pbn.LinearGaussianCPDType()

        assert type(cpd) == pbn.LinearGaussianCPD
        assert cpd.variable() == n
        assert set(cpd.evidence()) == set(spbn.parents(n))

    spbn.fit(df)
    
    spbn.remove_arc('a', 'b')

    cpd_b = spbn.cpd('b')
    assert type(cpd_b) == pbn.LinearGaussianCPD
    assert cpd_b.evidence != spbn.parents('b')

    spbn.fit(df)
    cpd_b = spbn.cpd('b')
    assert type(cpd_b) == pbn.LinearGaussianCPD
    assert cpd_b.evidence() == spbn.parents('b')

    spbn.set_node_type('c', pbn.CKDEType())

    with pytest.raises(ValueError) as ex:
        cpd_c = spbn.cpd('c')
    assert "not added" in str(ex.value)

    spbn.fit(df)
    cpd_c = spbn.cpd('c')
    assert cpd_c.type() == spbn.node_type('c')


def test_cpd():
    spbn = SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')], [('d', pbn.CKDEType())])

    with pytest.raises(ValueError) as ex:
        spbn.cpd('a')
    assert "not added" in str(ex.value)

    spbn.fit(df)

    assert spbn.cpd('a').type() == pbn.LinearGaussianCPDType()
    assert spbn.cpd('b').type() == pbn.LinearGaussianCPDType()
    assert spbn.cpd('c').type() == pbn.LinearGaussianCPDType()
    assert spbn.cpd('d').type() == pbn.CKDEType()

    assert spbn.cpd('a').fitted()
    assert spbn.cpd('b').fitted()
    assert spbn.cpd('c').fitted()
    assert spbn.cpd('d').fitted()

def test_add_cpds():
    spbn = SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')], [('d', pbn.CKDEType())])

    assert spbn.node_type('a') == pbn.UnknownFactorType()
    spbn.add_cpds([CKDE('a', [])])
    assert spbn.node_type('a') == pbn.CKDEType()

    with pytest.raises(ValueError) as ex:
        spbn.add_cpds([LinearGaussianCPD('d', ['a', 'b', 'c'])])
    assert "Bayesian network expects type" in str(ex.value)

    lg = LinearGaussianCPD('b', ['a'], [2.5, 1.65], 4)
    ckde = CKDE('d', ['a', 'b', 'c'])
    assert lg.fitted()
    assert not ckde.fitted()

    spbn.add_cpds([lg, ckde])

    spbn.set_node_type('a', pbn.UnknownFactorType())
    with pytest.raises(ValueError) as ex:
        not spbn.cpd('a').fitted()
    assert "CPD of variable \"a\" not added. Call add_cpds() or fit() to add the CPD." in str(ex.value)

    assert spbn.cpd('b').fitted()

    with pytest.raises(ValueError) as ex:
        not spbn.cpd('c').fitted()
    assert "CPD of variable \"c\" not added. Call add_cpds() or fit() to add the CPD." in str(ex.value)

    assert not spbn.cpd('d').fitted()

def test_logl():
    spbn = SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    spbn.fit(df)

    test_df = util_test.generate_normal_data(5000)
    ll = spbn.logl(test_df)
    sll = spbn.slogl(test_df)

    sum_ll = np.zeros((5000,))
    sum_sll = 0
    
    for n in spbn.nodes():
        cpd = spbn.cpd(n)
        l = cpd.logl(test_df)
        s = cpd.slogl(test_df)
        assert np.all(np.isclose(s, l.sum()))
        sum_ll += l
        sum_sll += s
    
    assert np.all(np.isclose(ll, sum_ll))
    assert np.isclose(sll, ll.sum())
    assert sll == sum_sll


import pytest
import numpy as np
from scipy.stats import gaussian_kde, norm
import pybnesian as pbn
import util_test

SIZE = 1000
df = util_test.generate_normal_data(SIZE)
seed = 0

def numpy_local_score(node_type, training_data, test_data, variable, evidence):
    if isinstance(variable, str):
        node_data = training_data.loc[:, [variable] + evidence].dropna()
        variable_data = node_data.loc[:, variable]
        evidence_data = node_data.loc[:, evidence]
        test_node_data = test_data.loc[:, [variable] + evidence].dropna()
        test_variable_data = test_node_data.loc[:, variable]
        test_evidence_data = test_node_data.loc[:, evidence]
    else:
        node_data = training_data.iloc[:, [variable] + evidence].dropna()
        variable_data = node_data.iloc[:, 0]
        evidence_data = node_data.iloc[:, 1:]
        test_node_data = test_data.iloc[:, [variable] + evidence].dropna()
        test_variable_data = test_node_data.iloc[:, 0]
        test_evidence_data = test_node_data.iloc[:, 1:]

    if node_type == pbn.LinearGaussianCPDType():
        N = variable_data.shape[0]
        d = evidence_data.shape[1]
        linregress_data = np.column_stack((np.ones(N), evidence_data.to_numpy()))
        (beta, res, _, _) = np.linalg.lstsq(linregress_data, variable_data.to_numpy(), rcond=None)
        var = res / (N - d - 1)

        means = beta[0] + np.sum(beta[1:]*test_evidence_data, axis=1)
        return norm.logpdf(test_variable_data, means, np.sqrt(var)).sum()
    elif node_type == pbn.CKDEType():
        k_joint = gaussian_kde(node_data.to_numpy().T,
                    bw_method=lambda s : np.power(4 / (s.d + 2), 1 / (s.d + 4)) * s.scotts_factor())
        if evidence:
            k_marg = gaussian_kde(evidence_data.to_numpy().T, bw_method=k_joint.covariance_factor())
            return np.sum(k_joint.logpdf(test_node_data.to_numpy().T) - k_marg.logpdf(test_evidence_data.to_numpy().T))
        else:
            return np.sum(k_joint.logpdf(test_node_data.to_numpy().T))

def test_holdout_create():
    s = pbn.HoldoutLikelihood(df)
    assert s.training_data().num_rows == 0.8 * SIZE
    assert s.test_data().num_rows == 0.2 * SIZE

    s = pbn.HoldoutLikelihood(df, 0.5)
    assert s.training_data().num_rows == 0.5 * SIZE
    assert s.test_data().num_rows == 0.5 * SIZE
    
    s = pbn.HoldoutLikelihood(df, 0.2, 0)
    s2 = pbn.HoldoutLikelihood(df, 0.2, 0)

    assert s.training_data().equals(s2.training_data())
    assert s.test_data().equals(s2.test_data())

    with pytest.raises(ValueError) as ex:
        s = pbn.HoldoutLikelihood(df, 10, 0)
    assert "test_ratio must be a number"  in str(ex.value)

    with pytest.raises(ValueError) as ex:
        s = pbn.HoldoutLikelihood(df, 0, 0)
    assert "test_ratio must be a number"  in str(ex.value)


def test_holdout_local_score_gbn():
    gbn = pbn.GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])
    
    hl = pbn.HoldoutLikelihood(df, 0.2, seed)

    assert np.isclose(hl.local_score(gbn, 'a', []), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'a', []))
    assert np.isclose(hl.local_score(gbn, 'b', ['a']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'b', ['a']))
    assert np.isclose(hl.local_score(gbn, 'c', ['a', 'b']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'c', ['a', 'b']))
    assert np.isclose(hl.local_score(gbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'd', ['a', 'b', 'c']))
    assert np.isclose(hl.local_score(gbn, 'd', ['a', 'b', 'c']), 
                      hl.local_score(gbn, 'd', ['b', 'c', 'a']))

    assert hl.local_score(gbn, 'a') == hl.local_score(gbn, 'a', gbn.parents('a'))
    assert hl.local_score(gbn, 'b') == hl.local_score(gbn, 'b', gbn.parents('b'))
    assert hl.local_score(gbn, 'c') == hl.local_score(gbn, 'c', gbn.parents('c'))
    assert hl.local_score(gbn, 'd') == hl.local_score(gbn, 'd', gbn.parents('d'))

def test_holdout_local_score_gbn_null():
    gbn = pbn.GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])
    
    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    hl = pbn.HoldoutLikelihood(df_null, 0.2, seed)

    assert np.isclose(hl.local_score(gbn, 'a', []), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'a', []))
    assert np.isclose(hl.local_score(gbn, 'b', ['a']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'b', ['a']))
    assert np.isclose(hl.local_score(gbn, 'c', ['a', 'b']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'c', ['a', 'b']))
    assert np.isclose(hl.local_score(gbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'd', ['a', 'b', 'c']))
    assert np.isclose(hl.local_score(gbn, 'd', ['a', 'b', 'c']), 
                      hl.local_score(gbn, 'd', ['b', 'c', 'a']))

    assert hl.local_score(gbn, 'a') == hl.local_score(gbn, 'a', gbn.parents('a'))
    assert hl.local_score(gbn, 'b') == hl.local_score(gbn, 'b', gbn.parents('b'))
    assert hl.local_score(gbn, 'c') == hl.local_score(gbn, 'c', gbn.parents('c'))
    assert hl.local_score(gbn, 'd') == hl.local_score(gbn, 'd', gbn.parents('d'))

def test_holdout_local_score_spbn():
    spbn = pbn.SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')],
                                [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    
    hl = pbn.HoldoutLikelihood(df, 0.2, seed)

    assert np.isclose(hl.local_score(spbn, 'a', []), 
                      numpy_local_score(pbn.CKDEType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'a', []))
    assert np.isclose(hl.local_score(spbn, 'b', ['a']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'b', ['a']))
    assert np.isclose(hl.local_score(spbn, 'c', ['a', 'b']), 
                      numpy_local_score(pbn.CKDEType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'c', ['a', 'b']))
    assert np.isclose(hl.local_score(spbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'd', ['a', 'b', 'c']))
    assert np.isclose(hl.local_score(spbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'd', ['b', 'c', 'a']))

    assert hl.local_score(spbn, 'a') == hl.local_score(spbn, 'a', spbn.parents('a'))
    assert hl.local_score(spbn, 'b') == hl.local_score(spbn, 'b', spbn.parents('b'))
    assert hl.local_score(spbn, 'c') == hl.local_score(spbn, 'c', spbn.parents('c'))
    assert hl.local_score(spbn, 'd') == hl.local_score(spbn, 'd', spbn.parents('d'))

def test_holdout_local_score_null_spbn():
    spbn = pbn.SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')],
                                [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])
    
    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    hl = pbn.HoldoutLikelihood(df_null, 0.2, seed)

    assert np.isclose(hl.local_score(spbn, 'a', []), 
                      numpy_local_score(pbn.CKDEType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'a', []))
    assert np.isclose(hl.local_score(spbn, 'b', ['a']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'b', ['a']))
    assert np.isclose(hl.local_score(spbn, 'c', ['a', 'b']), 
                      numpy_local_score(pbn.CKDEType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'c', ['a', 'b']))
    assert np.isclose(hl.local_score(spbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'd', ['a', 'b', 'c']))
    assert np.isclose(hl.local_score(spbn, 'd', ['a', 'b', 'c']), 
                      numpy_local_score(pbn.LinearGaussianCPDType(), hl.training_data().to_pandas(), hl.test_data().to_pandas(), 'd', ['b', 'c', 'a']))

    assert hl.local_score(spbn, 'a') == hl.local_score(spbn, 'a', spbn.parents('a'))
    assert hl.local_score(spbn, 'b') == hl.local_score(spbn, 'b', spbn.parents('b'))
    assert hl.local_score(spbn, 'c') == hl.local_score(spbn, 'c', spbn.parents('c'))
    assert hl.local_score(spbn, 'd') == hl.local_score(spbn, 'd', spbn.parents('d'))

def test_holdout_score():
    gbn = pbn.GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    hl = pbn.HoldoutLikelihood(df, 0.2, 0)

    assert np.isclose(hl.score(gbn), (
                            hl.local_score(gbn, 'a', []) +
                            hl.local_score(gbn, 'b', ['a']) +
                            hl.local_score(gbn, 'c', ['a', 'b']) +
                            hl.local_score(gbn, 'd', ['a', 'b', 'c'])))

    spbn = pbn.SemiparametricBN([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')], 
                                [('a', pbn.CKDEType()), ('c', pbn.CKDEType())])

    assert np.isclose(hl.score(spbn), (
                            hl.local_score(spbn, 'a') +
                            hl.local_score(spbn, 'b') +
                            hl.local_score(spbn, 'c') +
                            hl.local_score(spbn, 'd')))


from pybnesian import PartiallyDirectedGraph, MeekRules

def test_meek_rule1():
    # From Koller Chapter 3.4, Figure 3.12, pag 89.
    gr1 = PartiallyDirectedGraph(["X", "Y", "Z"], [("X", "Y")], [("Y", "Z")])

    assert MeekRules.rule1(gr1)
    assert gr1.num_edges() == 0
    assert set(gr1.arcs()) == set([('X', 'Y'), ('Y', 'Z')])

    assert not MeekRules.rule1(gr1)

def test_meek_rule2():
    # From Koller Chapter 3.4, Figure 3.12, pag 89.
    gr2 = PartiallyDirectedGraph(["X", "Y", "Z"], [("X", "Y"), ("Y", "Z")], [("X", "Z")])

    assert MeekRules.rule2(gr2)
    assert gr2.num_edges() == 0
    assert set(gr2.arcs()) == set([('X', 'Y'), ('Y', 'Z'), ('X', 'Z')])
    assert not MeekRules.rule2(gr2)

def test_meek_rule3():
    # From Koller Chapter 3.4, Figure 3.12, pag 89.
    gr3 = PartiallyDirectedGraph(["X", "Y1", "Y2", "Z"], [("Y1", "Z"), ("Y2", "Z")], [("X", "Y1"), ("X", "Y2"), ("X", "Z")])

    assert MeekRules.rule3(gr3)
    assert set(gr3.edges()) == set([('X', 'Y1'), ('X', 'Y2')])
    assert set(gr3.arcs()) == set([('X', 'Z'), ('Y1', 'Z'), ('Y2', 'Z')])
    assert not MeekRules.rule3(gr3)

def test_meek_sequential():
    # From Koller Chapter 3.4, Figure 3.13, pag 90.
    koller = PartiallyDirectedGraph(["A", "B", "C", "D", "E", "F", "G"], 
                                    [("B", "E"), ("C", "E")],
                                    [("A", "B"), ("B", "D"), ("C", "F"), ("E", "F"), ("F", "G")])
    changed = True
    while changed:
        changed = False
        changed = changed or MeekRules.rule1(koller)
        changed = changed or MeekRules.rule2(koller)
        changed = changed or MeekRules.rule3(koller)

    assert set(koller.edges()) == set([('A', 'B'), ('B', 'D')])
    assert set(koller.arcs()) == set([('B', 'E'), ('C', 'E'), ('E', 'F'), ('C', 'F'), ('F', 'G')])


import pytest
import numpy as np
import pandas as pd
import pyarrow as pa
import pybnesian as pbn
import util_test

df = util_test.generate_discrete_data_dependent(10000)

def test_data_type():
    a = pbn.DiscreteFactor("A", [])
    with pytest.raises(ValueError) as ex:
        a.data_type()
    "DiscreteFactor factor not fitted." in str(ex.value)
    
    categories = np.asarray(["a1", "a2"])
    a_values = pd.Categorical(categories[np.random.randint(len(categories), size=100)], categories=categories, ordered=False)
    df = pd.DataFrame({'A': a_values})
    a.fit(df)
    assert a.data_type() == pa.dictionary(pa.int8(), pa.string())

    categories = np.asarray(["a" + str(i) for i in range(1, 129)])
    a_values = pd.Categorical(categories[np.random.randint(len(categories), size=100)], categories=categories, ordered=False)
    df = pd.DataFrame({'A': a_values})
    a.fit(df)
    assert a.data_type() == pa.dictionary(pa.int8(), pa.string())

    categories = np.asarray(["a" + str(i) for i in range(1, 130)])
    a_values = pd.Categorical(categories[np.random.randint(len(categories), size=100)], categories=categories, ordered=False)
    df = pd.DataFrame({'A': a_values})
    a.fit(df)
    assert a.data_type() == pa.dictionary(pa.int16(), pa.string())

def test_fit():
    # a = DiscreteFactor('C', ['A', 'B'])
    a = pbn.DiscreteFactor('C', [])
    a.fit(df)



import numpy as np
import pybnesian as pbn
from pybnesian import BayesianNetworkType, BayesianNetwork
import util_test

df = util_test.generate_normal_data(1000)

def test_hc_estimate():
    bic = pbn.BIC(df)
    column_names = list(df.columns.values)
    start = pbn.GaussianNetwork(column_names)

    # Check algorithm with BN with nodes removed.
    column_names.insert(1, 'e')
    column_names.insert(3, 'f')
    start_removed_nodes = pbn.GaussianNetwork(column_names)
    start_removed_nodes.remove_node('e')
    start_removed_nodes.remove_node('f')

    arc_set = pbn.ArcOperatorSet()

    hc = pbn.GreedyHillClimbing()

    res = hc.estimate(arc_set, bic, start, max_iters=1)
    assert res.num_arcs() == 1
    added_arc = res.arcs()[0]
    op_delta = bic.score(res) - bic.score(start)

    res_removed = hc.estimate(arc_set, bic, start_removed_nodes, max_iters=1)
    assert res.num_arcs() == 1
    added_arc_removed = res_removed.arcs()[0]
    assert added_arc == added_arc_removed or added_arc == added_arc_removed[::-1]
    assert np.isclose(op_delta, bic.score(res_removed) - bic.score(start_removed_nodes))

    # BIC is score equivalent, so if we blacklist the added_arc, its reverse will be added.
    res = hc.estimate(arc_set, bic, start, max_iters=1, arc_blacklist=[added_arc])
    assert res.num_arcs() == 1
    reversed_arc = res.arcs()[0][::-1]
    assert added_arc == reversed_arc

    res_removed = hc.estimate(arc_set, bic, start_removed_nodes, max_iters=1, arc_blacklist=[added_arc_removed])
    assert res.num_arcs() == 1
    reversed_arc_removed = res_removed.arcs()[0][::-1]
    assert added_arc_removed == reversed_arc_removed

    assert np.isclose(op_delta, bic.local_score(res, added_arc[1], [added_arc[0]]) - bic.local_score(res, added_arc[1], []))
    assert np.isclose(op_delta, bic.local_score(res, added_arc_removed[1], [added_arc_removed[0]]) - bic.local_score(res, added_arc_removed[1], []))

    res = hc.estimate(arc_set, bic, start, epsilon=(op_delta + 0.01))
    assert res.num_arcs() == start.num_arcs()

    res_removed = hc.estimate(arc_set, bic, start_removed_nodes, epsilon=(op_delta + 0.01))
    assert res_removed.num_arcs() == start_removed_nodes.num_arcs()

    # Can't compare models because the arcs could be oriented in different direction, 
    # leading to a different search path. Execute the code, just to check no error is given.
    res = hc.estimate(arc_set, bic, start, verbose=False)
    res_removed = hc.estimate(arc_set, bic, start_removed_nodes, verbose=False)

def test_hc_conditional_estimate():
    bic = pbn.BIC(df)
    column_names = list(df.columns.values)

    start = pbn.ConditionalGaussianNetwork(column_names[2:], column_names[:2])
    
    nodes = column_names[2:]
    nodes.insert(1, 'e')
    interface_nodes = column_names[:2]
    interface_nodes.insert(1, 'f')
    start_removed_nodes = pbn.ConditionalGaussianNetwork(nodes, interface_nodes)
    start_removed_nodes.remove_node('e')
    start_removed_nodes.remove_interface_node('f')
    
    arc_set = pbn.ArcOperatorSet()
    hc = pbn.GreedyHillClimbing()

    res = hc.estimate(arc_set, bic, start, max_iters=1, verbose=False)
    assert res.num_arcs() == 1
    added_arc = res.arcs()[0]
    op_delta = bic.score(res) - bic.score(start)

    res_removed = hc.estimate(arc_set, bic, start_removed_nodes, max_iters=1, verbose=False)
    assert res_removed.num_arcs() == 1
    added_arc_removed = res_removed.arcs()[0]
    assert added_arc == added_arc_removed or added_arc == added_arc_removed[::-1]
    assert np.isclose(op_delta, bic.score(res_removed) - bic.score(start_removed_nodes))

    assert np.isclose(op_delta, bic.local_score(res, added_arc[1], [added_arc[0]]) -
                                bic.local_score(res, added_arc[1], []))
    assert np.isclose(op_delta, bic.local_score(res, added_arc_removed[1], [added_arc_removed[0]]) -
                                bic.local_score(res, added_arc_removed[1], []))

    res = hc.estimate(arc_set, bic, start, epsilon=(op_delta + 0.01))
    assert res.num_arcs() == start.num_arcs()
    res_removed = hc.estimate(arc_set, bic, start_removed_nodes, epsilon=(op_delta + 0.01))
    assert res_removed.num_arcs() == start_removed_nodes.num_arcs()

    res = hc.estimate(arc_set, bic, start, verbose=False)
    assert all(map(lambda arc : not res.is_interface(arc[1]), res.arcs()))
    res_removed = hc.estimate(arc_set, bic, start_removed_nodes, verbose=False)
    assert all(map(lambda arc : not res_removed.is_interface(arc[1]), res_removed.arcs()))

def test_hc_estimate_validation():
    column_names = list(df.columns.values)
    start = pbn.GaussianNetwork(column_names)

    column_names.insert(1, 'e')
    column_names.insert(4, 'f')
    start_removed_nodes = pbn.GaussianNetwork(column_names)
    start_removed_nodes.remove_node('e')
    start_removed_nodes.remove_node('f')
    
    vl = pbn.ValidatedLikelihood(df)
    arc_set = pbn.ArcOperatorSet()

    hc = pbn.GreedyHillClimbing()

    res = hc.estimate(arc_set, vl, start, max_iters=1)
    assert res.num_arcs() == 1
    added_arc = res.arcs()[0]
    op_delta = vl.cv_lik.score(res) - vl.cv_lik.score(start)

    res_removed = hc.estimate(arc_set, vl, start_removed_nodes, max_iters=1)
    assert res_removed.num_arcs() == 1
    added_arc_removed = res_removed.arcs()[0]
    assert added_arc == added_arc_removed or added_arc == added_arc_removed[::-1]
    assert np.isclose(op_delta, vl.cv_lik.score(res_removed) - vl.cv_lik.score(start_removed_nodes))

    assert np.isclose(op_delta, vl.cv_lik.local_score(res, added_arc[1], [added_arc[0]]) - 
                                vl.cv_lik.local_score(res, added_arc[1], []))
    assert np.isclose(op_delta, vl.cv_lik.local_score(res, added_arc_removed[1], [added_arc_removed[0]]) -
                                vl.cv_lik.local_score(res, added_arc_removed[1], []))

    # CV is score equivalent for GBNs, so if we blacklist the added_edge, its reverse will be added.
    res = hc.estimate(arc_set, vl, start, max_iters=1, arc_blacklist=[added_arc])
    assert res.num_arcs() == 1
    reversed_arc = res.arcs()[0][::-1]
    assert added_arc == reversed_arc

    res_removed = hc.estimate(arc_set, vl, start_removed_nodes, max_iters=1, arc_blacklist=[added_arc_removed])
    assert res_removed.num_arcs() == 1
    reversed_arc_removed = res_removed.arcs()[0][::-1]
    assert reversed_arc == reversed_arc_removed
    
    res = hc.estimate(arc_set, vl, start, epsilon=(op_delta + 0.01))
    assert res.num_arcs() == start.num_arcs()

    res_removed = hc.estimate(arc_set, vl, start_removed_nodes, epsilon=(op_delta + 0.01))
    assert res_removed.num_arcs() == start_removed_nodes.num_arcs()

    # Can't compare models because the arcs could be oriented in different direction, 
    # leading to a different search path. Execute the code, just to check no error is given.
    res = hc.estimate(arc_set, vl, start, verbose=False)
    res_removed = hc.estimate(arc_set, vl, start_removed_nodes, verbose=False)

def test_hc_shortcut_function():
    model = pbn.hc(df, bn_type=pbn.GaussianNetworkType())
    assert type(model) == pbn.GaussianNetwork

    model = pbn.hc(df, bn_type=MyRestrictedGaussianNetworkType(), score="bic", operators=["arcs"])
    assert type(model) == NewBN

class MyRestrictedGaussianNetworkType(BayesianNetworkType):
    def __init__(self):
        BayesianNetworkType.__init__(self)

    def is_homogeneous(self):
        return True

    def default_node_type(self):
        return pbn.LinearGaussianCPDType()

    def can_have_arc(self, model, source, target):
        return "a" in source

    def new_bn(self, nodes):
        return NewBN(nodes)

    def __str__(self):
        return "MyRestrictedGaussianNetworkType"

class NewBN(BayesianNetwork):
    def __init__(self, variables, arcs=None):
        if arcs is None:
            BayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables)
        else:
            BayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, arcs)

        self.extra_data = "extra"

    def __getstate_extra__(self):
        return self.extra_data

    def __setstate_extra__(self, extra):
        self.extra_data = extra

def test_newbn_estimate_validation():
    start = NewBN(["a", "b", "c", "d"])
    hc = pbn.GreedyHillClimbing()
    arc = pbn.ArcOperatorSet()
    bic = pbn.BIC(df)

    estimated = hc.estimate(arc, bic, start)

    assert type(start) == type(estimated)
    assert estimated.extra_data == "extra"